#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The launcher for EAvatar to run in a text console or headless environment.
"""

import sys
import logging
import multiprocessing

#makes multiprocessing work when in freeze mode.
multiprocessing.freeze_support()



# prevent no handler warning
try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logger = logging.getLogger("ava")


def main():
    from ava.cmds.cli import cli, load_commands

    load_commands()
    return cli(auto_envvar_prefix=b'AVA')

if __name__ == '__main__':
    sys.exit(main())


# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import os
import unittest
import gevent

from ava.spi.defines import AVA_USER_XID, AVA_AGENT_SECRET
from ava.runtime import settings
from ava.core.agent import Agent


class AgentTest(unittest.TestCase):
    """
    For functional tests which require a running agent.
    """
    agent = None
    user_xid = b'AYPwK3c3VK7ZdBvKfcbV5EmmCZ8zSb9viZ288gKFBFuE92jE'
    user_key = b'Kd2xqKsjTnhhqXjY64eeSEyS1i9kSGTHt9S57sqeK51bkPRh'
    user_secret = b'SVQh1mgbdvuFoZihYH8urZyBGpfZ4PJnn8af2R9MuqZyktHa'

    agent_secret = b'SYNmgyQqhAnVwKLrmSmYzahkzH3V51qdShL41JFPnmsZob96'

    @classmethod
    def setUpClass(cls):
        settings['debug'] = True
        os.environ.setdefault(AVA_USER_XID, cls.user_xid)
        os.environ.setdefault(AVA_AGENT_SECRET, cls.agent_secret)
        AgentTest.agent = Agent()
        agent_greenlet = gevent.spawn(AgentTest.agent.run)
        while not AgentTest.agent.running:
            gevent.sleep(0.5)

    @classmethod
    def tearDownClass(cls):
        AgentTest.agent.interrupted = True
        while AgentTest.agent.running:
            gevent.sleep(0.5)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import requests
from ava.util.tests import AgentTest


class ApiTest(AgentTest):

    def setUp(self):
        self.base_url = 'http://127.0.0.1:5080'

    #### Root resource ####
    def test_get_root_resource(self):
        r = requests.get(self.base_url)
        self.assertEqual(r.status_code, 200)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import sys

def create():
    """
    Creates the shell based on a specified platform.

    :param platform: The platform identifier.
    :return:
    """
    plat = sys.platform
    if plat.startswith("win32"):
        from ava.shell.win32 import Shell
    elif plat.startswith("darwin"):
        from ava.shell.osx import Shell
    elif plat.startswith("linux"):
        from ava.shell.gtk import Shell
    else:
        from ava.shell.console import Shell

    return Shell()


# -*- coding: utf-8 -*-

"""
The module contains global settings and variables for the application.
"""

from __future__ import absolute_import, division, print_function

from .environ import (
    base_dir, pod_dir, pkgs_dir, logs_dir, conf_dir, data_dir, mods_dir)

from .config import settings



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
"""
Commands for controlling the agent process.
"""

import sys
import gevent
import bottle
from ava.spi.webfront import dispatcher, check_authentication

api = dispatcher.mount(b'agent', bottle.Bottle())
api.add_hook('before_request', check_authentication)


def _restart():
    """
    Tells shell to restart the agent process.
    """
    sys.exit(1)


@api.post("/restart")
def restart():
    gevent.spawn_later(3, _restart)
    return "Agent is restarting..."


@api.get("/status")
def status():
    return "Agent is running..."





# -*- coding: utf-8 -*-
"""
Task declaration
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from .context import instance as agent

_engine = agent().task_engine


def task(func):
    """
    Marks a function as a task template(code, arguments, etc).

    :param func:
    :return: the task wrapping given function object.
    """
    return _engine.register(func)


def run_once(task, seconds=0, args=[], kwargs={}):
    """
    Run a one-time task later after the specified seconds.

    :param task: the task proxy
    :param seconds: the delayed seconds before running.
    :return: the schedule.
    """
    return task.run_once(seconds, args, kwargs)


def run_periodic(task, interval, start_time=None, stop_time=None,
                 args=[], kwargs={}):
    """

    :param task: the task proxy.
    :param interval: the interval in seconds.
    :param start_time: the timestamp from which the task can be run
    :param stop_time: the timestamp before which the task should be run
    :return: the schedule.
    """
    return task.run_periodic(interval, start_time, stop_time,
                                args, kwargs)


def cancel_schedule(sched):
    _engine.cancel(sched)


def cancel_schedule_by_id(sched_id):
    sched = _engine.get_schedule(sched_id)
    if sched:
        _engine.cancel(sched)


def get_schedule_by_id(sched_id):
    """ Gets schedule by ID.

    :param sched_id:
    :return:
    """
    return _engine.get_schedule(sched_id)


# -*- coding: utf-8 -*-
"""
Various encoding/decoding helpers.
"""

from __future__ import print_function, division, absolute_import

import base64


def base64url_encode(src):
    """
    URL-safe base64 encoding without padding.

    :param src:
    :return:
    """
    return base64.urlsafe_b64encode(src).replace('=','')


def base64url_decode(src):
    """
    Decode URL-safe base64-encoded string without padding.

    :param src:
    :return:
    """
    return base64.urlsafe_b64decode(src + '=' * (len(src) % 4 ))


# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
import os
import json
import click
from ava.util import crypto
from ava.util import yaml
from ava.runtime import environ
from .cli import cli


def gen_random_key(ctx):
    if ctx.obj['verbosity']:
        click.echo("Generating random key...")
    (pk, sk) = crypto.generate_keypair()
    xid = crypto.key_to_xid(pk)
    secret_str = crypto.secret_to_string(sk)
    key_str = crypto.key_to_string(pk)
    res = dict(xid=xid, key=key_str, secret=secret_str)
    click.echo(json.dumps(res, sort_keys=True, indent=4, separators=(',', ': ')))
    return 0


@cli.group()
@click.pass_context
def key(ctx):
    """ Key management
    """
    pass


@key.command()
@click.option('--email', prompt='Your email address', help='Email address')
@click.option('--password',
              prompt='Your password',
              hide_input=True,
              confirmation_prompt=True,
              help='Your password')
@click.option('--force', '-f',
              help='Force to generate new keys even if already exists.')
@click.pass_context
def user(ctx, email, password, force=False):
    """ Set up user keys.
    """
    filepath = os.path.join(ctx.obj['app_dir'], 'user-keys.yml')
    if os.path.exists(filepath) and not force:
        click.echo('Keys already exist, use --force option to replace.',
                   err=True)
        click.confirm('Do you want to overwrite it?', abort=True)

    content = {}
    secret = crypto.derive_secret_key(password=password.encode('utf-8'),
                                      salt=email.encode('utf-8'))
    content[b'xid'] = crypto.secret_key_to_xid(secret)
    content[b'email'] = email

    oldmask = os.umask(077)
    with open(filepath, 'wb') as keyfile:
        yaml.dump(content, keyfile)
    os.umask(oldmask)
    if ctx.obj['verbosity'] > 0:
        click.echo('Key file generated.')


@key.command()
@click.option('--force', '-f',
              help='Force to generate new keys even if already exists.')
@click.pass_context
def agent(ctx, force=False):
    """ Generate keys for agent.
    """
    userkeys_path = os.path.join(ctx.obj['app_dir'], 'user-keys.yml')
    if not os.path.exists(userkeys_path):
        click.echo('You should set user keys first.', err=True)
        return

    with open(userkeys_path, 'rb') as userkeys_file:
        userkeys = yaml.load(userkeys_file)
    user_xid = userkeys.get('xid')

    agentkeys_path = os.path.join(environ.conf_dir(), 'ava-keys.yml')
    if os.path.exists(agentkeys_path) and not force:
        click.echo('Agent keys exist!')
        click.confirm('Do you want to overwrite it?', abort=True)

    content = {}
    pk, sk = crypto.generate_keypair()

    content[b'user'] = user_xid
    content[b'secret'] = crypto.secret_to_string(sk)

    with open(agentkeys_path, 'wb') as agentkeys_file:
        yaml.dump(content, agentkeys_file)


@key.command()
@click.option('--salt', '-s',
              help='Value for increasing randomness, e.g. email address.')
@click.option('--password', '-p',
              help='Hard to guess password.')
@click.pass_context
def generate(ctx, salt=None, password=None):
    """ Generate random key or derive from salt and password.

    """
    if not salt and not password:
        return gen_random_key(ctx)

    if not (salt and password):
        if not salt:
            click.echo("Missing option: salt")
        if not password:
            click.echo("Missing option: password")
        return 1

    sk = crypto.derive_secret_key(password, salt)
    (pk, sk) = crypto.generate_keypair(sk=sk)
    xid = crypto.key_to_xid(pk)
    secret_str = crypto.secret_to_string(sk)
    key_str = crypto.key_to_string(pk)
    res = dict(xid=xid, key=key_str, secret=secret_str)
    click.echo(json.dumps(res, sort_keys=True, indent=4, separators=(',', ': ')))
    return 0


@key.command()
@click.option('--xid', '-i',
              help='Exchange ID')
@click.option('--key', '-k',
              help='Public key')
@click.option('--secret', '-s',
              help='Secret key')
@click.pass_context
def validate(ctx, xid, key, secret):
    """
    Validate XID, public key or secret key.

    :param ctx:
    :param xid:
    :param key:
    :param secret:
    :return:
    """
    if xid:
        if crypto.validate_xid(xid):
            click.echo("XID %s is valid." % xid)
        else:
            click.echo("XID %s is invalid." % xid)

    if key:
        if crypto.validate_key_string(key):
            click.echo("Key is valid")
        else:
            click.echo("Key is invalid.")

    if secret:
        if crypto.validate_secret_string(secret):
            click.echo("Secret is valid.")
        else:
            click.echo("Secret is invalid.")

    if key and secret:
        sk = crypto.string_to_secret(secret)
        pk = crypto.string_to_key(key)
        (vpk, vsk) = crypto.generate_keypair(sk=sk)
        if vpk == pk and vsk == sk:
            click.echo("Key and secret are matched.")
            return 0
        click.echo("Key and secret are mismatched.")
        return 1




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


import logging
import gevent
from uuid import uuid1
from gevent import Greenlet
from datetime import datetime
from ava.spi.errors import TaskNotRegistered, TaskAlreadyRegistered

logger = logging.getLogger(__name__)


class TaskProxy(object):
    def __init__(self, task_engine, func):
        self.task_engine = task_engine
        self.func = func
        self.key = func.__module__ + '.' + func.func_name

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def run_once(self, delayed_secs=0, args=[], kwargs={}):
        return self.task_engine.run_once(self.key, delayed_secs, args, kwargs)

    def run_periodic(self, interval,
                     start_time=None, stop_time=None,
                     args=[], kwargs={}):
        return self.task_engine.run_periodic(self.key, interval,
                                             start_time=None, stop_time=None,
                                             args=[], kwargs={})


class Schedule(Greenlet):
    def __init__(self, sched_id, task, args=[], kwargs={}):
        super(Schedule, self).__init__()
        self.id = sched_id
        self.task = task
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.error = None

    def call(self):
        try:
            logger.debug("Before running task:")
            self.result = self.task(*self.args, **self.kwargs)
            logger.debug("Task result: %r", self.result)
            return self.result
        except Exception as ex:
            logger.error("Error in calling task: %s", self.task.key)
            self.error = ex

    def _run(self):
        raise NotImplementedError()


class OnceSchedule(Schedule):
    def __init__(self, sched_id, task, seconds=0, args=[], kwargs={}):
        super(OnceSchedule, self).__init__(sched_id, task, args, kwargs)
        self.seconds = seconds

    def _run(self):
        if self.seconds > 0:
            gevent.sleep(self.seconds)
        self.call()


class PeriodicSchedule(Schedule):
    def __init__(self, sched_id, task, interval, start_time=None,
                 stop_time=None, args=[], kwargs={}):
        super(PeriodicSchedule, self).__init__(sched_id, task, args, kwargs)
        self.interval = interval
        self.start_time = start_time
        self.stop_time = stop_time
        self.next_run = 0

    def _run(self):
        now = datetime.now()
        if self.start_time is None:
            self.start_time = now

        if now < self.start_time:
            self.next_run = self.start_time - now
            gevent.sleep(self.next_run)

        while True:
            self.call()
            now = datetime.now()
            if self.stop_time and now >= self.stop_time:
                break
            self.next_run = self.interval
            gevent.sleep(self.next_run)


class TaskEngine(object):

    def __init__(self):
        self.context = None
        self._schedules = {}
        self._tasks = {}

    def start(self, ctx):
        logger.debug("Starting task engine...")
        self.context = ctx
        self.context['task_engine'] = self

    def stop(self, ctx):
        logger.debug("Stopping task engine...")
        for sched in self._schedules.values():
            sched.kill()
        gevent.joinall(self._schedules.values())

    def register(self, func):
        task_key = func.__module__ + '.' + func.func_name
        if self._tasks.get(task_key) is not None:
            raise TaskAlreadyRegistered(task_key)

        proxy = TaskProxy(self, func)
        self._tasks[task_key] = proxy
        return proxy

    def unregister(self, task_key):
        proxy = self._tasks.get(task_key)
        if proxy is not None:
            del self._tasks[task_key]

    def get_task(self, task_key):
        return self._tasks.get(task_key)

    def run_once(self, task_key, delayed_secs, args, kwargs):
        """
        Schedules a one-time task.

        :param task_key:
        :param delayed_secs:
        :return: the schedule
        """

        task = self._tasks.get(task_key)
        if task is None:
            raise TaskNotRegistered(task_key)
        schedule_id = uuid1().hex
        schedule = OnceSchedule(schedule_id, task, delayed_secs, args, kwargs)
        self._schedules[schedule_id] = schedule
        schedule.start()
        return schedule

    def run_periodic(self, task_key, interval,
                     start_time=None, stop_time=None,
                     args=[], kwargs={}):
        """
        Schedules a periodic task.

        :param task_key:
        :param interval:
        :param start_time: If None, start immediately.
        :param stop_time: If None, the task run indefinitely.
        :return: the schedule
        """
        task = self._tasks.get(task_key)
        if task is None:
            raise TaskNotRegistered()

        schedule_id = uuid1().hex
        schedule = PeriodicSchedule(schedule_id, task, interval,
                                    start_time, stop_time,
                                    args, kwargs)
        self._schedules[schedule_id] = schedule
        schedule.start()
        return schedule

    def cancel(self, schedule):
        """
        Cancels the scheduled task.

        :param schedule_id:
        :return:
        """
        if schedule is None:
            return False
        del self._schedules[schedule.id]
        schedule.kill()
        return True

    def get_schedule(self, sched_id):
        """ Gets the schedule via the given id.

        :param sched_id: the schedule id.
        :return:
        """
        return self._schedules.get(sched_id)


# -*- coding: utf-8 -*-
"""
The module provides functions for access token manipulation.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import binascii
import logging
from hashlib import sha256
from calendar import timegm
from collections import Mapping
from datetime import datetime, timedelta

from .crypto import (
    string_to_key,
    string_to_secret,
    xid_to_key,
    public_key_encrypt,
    public_key_decrypt
)

from .codecs import base64url_decode, base64url_encode
from .compat import text_type, string_types

logger = logging.getLogger(__name__)


def timedelta_total_seconds(delta):
    try:
        delta.total_seconds
    except AttributeError:
        # On Python 2.6, timedelta instances do not have
        # a .total_seconds() method.
        total_seconds = delta.days * 24 * 60 * 60 + delta.seconds
    else:
        total_seconds = delta.total_seconds()

    return total_seconds


class DecodeError(ValueError):
    pass


class ExpiredSignatureError(ValueError):
    pass


def sign(msg, issuer_sk, audience_pk):
    if isinstance(issuer_sk, string_types):
        logger.debug("issuer_sk: %s", issuer_sk)
        issuer_sk = string_to_secret(issuer_sk)
    if isinstance(audience_pk, string_types):
        logger.debug("audience_pk: %s", audience_pk)
        audience_pk = string_to_key(audience_pk)

    h = sha256(msg).digest()
    sig = public_key_encrypt(issuer_sk, audience_pk, h)
    return sig


def _verify(msg, issuer_pk, audience_sk, sig):
#    if isinstance(issuer_pk, text_type):
#        issuer_pk = string_to_key(issuer_pk)
    if isinstance(audience_sk, string_types):
        logger.debug("audience_sk: %s", audience_sk)
        audience_sk = string_to_secret(audience_sk)

    hashval = sha256(msg).digest()
    verifyval = public_key_decrypt(audience_sk, issuer_pk, sig)
    return hashval == verifyval


def encode(payload, issuer_sk, audience_pk, headers=None):
    segments = []
    header = {'typ': 'JWT', 'alg': 'EA256'}
    if headers:
        header.update(headers)

    json_header = json.dumps(
        header,
        separators=(',', ':')
    ).encode('utf-8')

    segments.append(base64url_encode(json_header))

    # Payload
    for time_claim in ['exp', 'iat', 'nbf']:
        # Convert datetime to a intDate value in known time-format claims
        if isinstance(payload.get(time_claim), datetime):
            payload[time_claim] = timegm(payload[time_claim].utctimetuple())

    json_payload = json.dumps(
        payload,
        separators=(',', ':')
    ).encode('utf-8')

    segments.append(base64url_encode(json_payload))

    # Segments
    signing_input = b'.'.join(segments)
    signature = sign(signing_input, issuer_sk, audience_pk)

    segments.append(base64url_encode(signature))

    return b'.'.join(segments)


def decode(tok, audience_sk, verify=True, **kwargs):
    payload, signing_input, header, signature = load(tok)
    if verify:
        verify_signature(payload, signing_input, header, signature, audience_sk, **kwargs)

    return payload


def verify_signature(payload, signing_input, header, signature, audience_sk,
                     verify_expiration=True, leeway=0, audience=None):
    logger.debug("audience_sk: %s", audience_sk)

    if isinstance(leeway, timedelta):
        leeway = timedelta_total_seconds(leeway)

    issuer_pk = payload.get('iss')
    if issuer_pk is None:
        issuer_pk = payload.get('sub')
    if issuer_pk is None:
        raise DecodeError("Issuer or subject not found.")
    issuer_pk = xid_to_key(issuer_pk)

    if not _verify(signing_input, issuer_pk, audience_sk, signature):
        raise DecodeError('Signature verification failed')

    if 'nbf' in payload and verify_expiration:
        utc_timestamp = timegm(datetime.utcnow().utctimetuple())

        if payload['nbf'] > (utc_timestamp + leeway):
            raise ExpiredSignatureError('Signature not yet valid')

    if 'exp' in payload and verify_expiration:
        utc_timestamp = timegm(datetime.utcnow().utctimetuple())

        if payload['exp'] < (utc_timestamp - leeway):
            raise ExpiredSignatureError('Signature has expired')


def load(tok):
    if isinstance(tok, text_type):
        tok = tok.encode('utf-8')
    try:
        signing_input, crypto_segment = tok.rsplit(b'.', 1)
        header_segment, payload_segment = signing_input.split(b'.', 1)
    except ValueError:
        raise DecodeError('Not enough segments')

    try:
        header_data = base64url_decode(header_segment)
    except (TypeError, binascii.Error):
        raise DecodeError('Invalid header padding')
    try:
        header = json.loads(header_data.decode('utf-8'))
    except ValueError as e:
        raise DecodeError('Invalid header string: %s' % e)
    if not isinstance(header, Mapping):
        raise DecodeError('Invalid header string: must be a json object')

    try:
        payload_data = base64url_decode(payload_segment)
    except (TypeError, binascii.Error):
        raise DecodeError('Invalid payload padding')
    try:
        payload = json.loads(payload_data.decode('utf-8'))
    except ValueError as e:
        raise DecodeError('Invalid payload string: %s' % e)
    if not isinstance(payload, Mapping):
        raise DecodeError('Invalid payload string: must be a json object')

    try:
        signature = base64url_decode(crypto_segment)
    except (TypeError, binascii.Error):
        raise DecodeError('Invalid crypto padding')

    return (payload, signing_input, header, signature)


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
"""
Loading modules that provide task codes.
"""
import os
import sys
import glob
import logging
from importlib import import_module
from ava.runtime import environ
from ava.spi.signals import MODULE_LOADED

logger = logging.getLogger(__name__)

_MODULES_DIR = os.path.join('mods', 'tasks')

# the package name for modules.
_MODULE_PKG = 'mods.tasks.'


class ModuleInfo(object):

    def __init__(self, name, mod):
        self._name = name
        self._mod = mod

    @property
    def name(self):
        return self._name

    @property
    def module(self):
        return self._mod

    @module.setter
    def module(self, mod):
        self._mod = mod


class TaskModEngine(object):
    """
    Responsible for managing application modules.
    """

    def __init__(self):
        self.modules = {}
        self.modules_path = os.path.join(environ.pod_dir(), _MODULES_DIR)
        self.modules_path = os.path.abspath(self.modules_path)

    def _scan_modules(self):
        pattern = os.path.join(self.modules_path, '[a-zA-Z][a-zA-Z0-9_]*.py')
        return glob.glob(pattern)

    def _load_modules(self, ctx):
        sys.path.append(environ.pod_dir())
        logger.debug("Tasks module directory: %s", self.modules_path)

        module_files = self._scan_modules()

        logger.debug("Found %d task module(s)" % len(module_files))

        for s in module_files:
            name = os.path.basename(s)
            if '__init__.py' == name:
                continue

            # gets the basename without extension part.
            name = os.path.splitext(name)[0]
            try:
                logger.debug("Loading task module: %s", name)
                mod = import_module(_MODULE_PKG + name)
                mod_info = ModuleInfo(name, mod)
                self.modules[name] = mod_info

                ctx.send(signal=MODULE_LOADED, sender=self)
            except ImportError:
                logger.error("Failed to import task module: %s", name, exc_info=True)

    def start(self, ctx):
        logger.debug("Starting task module engine...")
        self._load_modules(ctx)
        logger.debug("Task module engine started.")

    def stop(self, ctx):
        logger.debug("Task module engine stopped.")


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
"""
Intended to provide the definitions of significant signals.
"""

GREENLET_CREATED = 'greenlet.created'

AGENT_STARTED = "agent.started"
AGENT_STOPPING = "agent.stopping"

MODULE_LOADED = "module.loaded"
MODULE_UNLOADED = "module.unloaded"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import click


from .cli import cli

logger = logging.getLogger(__name__)


@cli.command()
@click.pass_context
def run(ctx):
    """ Run the agent

    """
    from ava.shell.console import Shell
    if ctx.obj['verbosity'] > 1:
        logger.debug("Starting the shell...")

    shell = Shell()
    shell.do_run()

    if ctx.obj['verbosity'] > 1:
        logger.debug("Shell stopped.")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import webbrowser
import multiprocessing
import logging
from abc import ABCMeta, abstractmethod

from ava.runtime import settings

STR_STATUS = 'Ava - running...'
STR_OPEN_WEBFRONT = u'Open Webfront'
STR_EXIT = u'Quit Ava'

logger = logging.getLogger(__name__)


def start_server_process():
    """
    Starts the server in the forked process.
    :return: None
    """
    from ava.core.agent import start_agent
    start_agent()


class Supervisor(object):
    """
    Responsible for monitor and manage the server process.
    """
    def __init__(self, target=start_server_process):
        self.server_process = None
        self.restarted_times = 0
        self.target = target

    def start_server(self):
        if self.is_server_running():
            return

        self.server_process = multiprocessing.Process(target=self.target)
        self.server_process.start()

    def stop_server(self):
        if self.server_process is not None:
            self.server_process.terminate()
            self.server_process.join()

        self.server_process = None

    def is_server_running(self):
        return self.server_process is not None

    def health_check(self):
        """
        :return True if everything is OK.
        """

        if not self.is_server_running():
            logger.debug("Server is not running, no check.")
            return

        #logger.debug("Doing check.")
        exitcode = self.server_process.exitcode

        if exitcode is None:
            return True

        if exitcode < 0 or exitcode == 1:
            if self.restarted_times > 5:
                logger.error("Server restarted more than 5 times, give up!")
                self.server_process = None
                return False

            self.server_process = multiprocessing.Process(target=start_server_process)
            self.server_process.start()
            self.restarted_times += 1
            logger.warning("Server process restarted.")
            return True
        else:
            logger.info("Server process exited.")
            self.server_process = None
            return False


class ShellBase(object):
    """
    Base class for Shell implementations.
    Shell is responsible for launching server process and provides machinery for the user to stop it.
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        port = settings['webfront']['listen_port']
        self.base_url = "http://127.0.0.1:%d/" % (port,)
        self.supervisor = Supervisor()
        self.shell_stopped = False

    def open_main_ui(self):
        webbrowser.open(self.base_url)

    def is_server_running(self):
        return self.supervisor.is_server_running()

    def set_base_url(self, url):
        self.base_url = url

    def start_server(self):
        self.supervisor.start_server()

    def stop_server(self):
        self.supervisor.stop_server()

    def check_server(self):
        """
        Checks if server process in running.
        """
        #logger.debug("Checking server status...")

        if not self.supervisor.health_check():
            self.shell_stopped = True

    def _on_idle(self):
        """
        Subclass should run this method from time to time.
        """
        self.check_server()

    def do_run(self):
        self.start_server()

        try:
            self.run()
            logger.debug("Shell is stopping.")
        finally:
            logger.debug("Stopping server")
            self.stop_server()

    @abstractmethod
    def run(self):
        """
        Subclass must implement this method to launch the shell.
        """





# -*- coding: utf-8 -*-
"""
Various definitions used across different packages.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from ava import VERSION_MAJOR, VERSION_MINOR, VERSION_MICRO, VERSION_STRING


# return as the root resource.
AGENT_INFO = {
    "EAvatar": "A versatile agent.",
    "version": VERSION_STRING,
    "vendor": {
        "name": "EAvatar Technology Ltd.",
        "version": "0.1.0"
    },
}


# activated engines

INSTALLED_ENGINES = [
    "ava.core.data:DataEngine",
    "ava.core.webfront:WebfrontEngine",
    "ava.core.task:TaskEngine",
    "ava.core.extension:ExtensionEngine",
    "ava.core.mod_tasks:TaskModEngine",
    "ava.core.mod_webhooks:WebhooksEngine",
#    "ava.core.websocket:WebsocketEngine",
]


##### Environment variable ####
AVA_POD_FOLDER = 'AVA_POD'  # where the working directory.
AVA_AGENT_SECRET = 'AVA_AGENT_SECRET'  # secret key for this agent.
AVA_USER_XID = 'AVA_USER_XID'  # the owner's public key.


# tries to import definitions from the global settings.

try:
    from settings import *
except ImportError:
    pass


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from pydispatch import dispatcher


class Context(object):
    """
    The top context for all other entities.
    """
    def __init__(self, agent):
        self._agent = agent
        self._attributes = {}
        self._dispatcher = dispatcher

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        self.bind(key, value)

    def __delitem__(self, key):
        self.unbind(key)

    def __getattr__(self, item):
        return self.get(item)

    def bind(self, key, provider):
        self._attributes[key] = provider

    def unbind(self, key):
        del self._attributes[key]

    def get(self, key, default=None):
        obj = self._attributes.get(key)
        if not obj:
            return default
        elif callable(obj):
            return obj()
        else:
            return obj

    def add_child_greenlet(self, child):
        self._agent.add_child_greenlet(child)

    def send(self, *args, **kwargs):
        """
        Send signal/event to registered receivers.

        :param args:
        :param kwargs:
        :return:
        """
        self._dispatcher.send(*args, **kwargs)

    def connect(self, receiver, *args, **kwargs):
        """
        Connect the receiver to listen for signals/events.
        :param signal:
        :param sender:
        :return:
        """
        self._dispatcher.connect(receiver, *args, **kwargs)

    def disconnect(self, receiver, *args, **kwargs):
        """
        Disconnect the specified receiver.
        :return:
        """
        self._dispatcher.disconnect(receiver, *args, **kwargs)


class Component(object):
    """
    Base for component classes.
    """
    pass


####  Singleton construction  ####
_context = None


def instance(agent=None):
    """
    Gets the context singleton. Must first be invoked by agent.

    :param agent:
    :return:
    """
    global _context
    if _context is None:
        _context = Context(agent)

    return _context



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from gevent import monkey
monkey.patch_all()

import unittest
import gevent
import ws4py
from ws4py.client.geventclient import WebSocketClient


class WebsocketTest(unittest.TestCase):

    def test_websocket_connection(self):
        ws = WebSocketClient('ws://127.0.0.1:5080/ws')
        ws.connect()
        print("Sending: 'Hello, World'...")
        ws.send("Hello, World")
        print("Sent")
        print("Receiving...")
        result = ws.receive()
        print("Received: {}".format(result))
        ws.close()
        self.assertEqual("Hello, World", str(result))

    def test_secure_websocket_connection(self):
        ws = WebSocketClient('wss://127.0.0.1:5443/ws')
        ws.connect()
        print("Sending: 'Hello, World'...")
        ws.send("Hello, World")
        print("Sent")
        print("Receiving...")
        result = ws.receive()
        print("Received: {}".format(result))
        ws.close()
        self.assertEqual("Hello, World", str(result))


# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from .webfront import dispatcher
from ws4py.websocket import EchoWebSocket
from ws4py.server.wsgiutils import WebSocketWSGIApplication


class WebSocketHandler(EchoWebSocket):
    def __init__(self, *args, **kwargs):
        super(WebSocketHandler, self).__init__(*args, **kwargs)


class WebsocketEngine(object):
    def start(self, ctx=None):
        dispatcher.mount('/ws', WebSocketWSGIApplication(
            handler_cls=WebSocketHandler))

    def stop(self, ctx=None):
        pass


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import os
import unittest

from ava.runtime import environ


class RuntimeEnvironTest(unittest.TestCase):

    def test_work_home(self):
        env = environ.Environment()
        self.assertTrue(os.path.isdir(env.pod_dir))
        self.assertTrue(os.path.isdir(env.conf_dir))
        self.assertTrue(os.path.isdir(env.data_dir))
        self.assertTrue(os.path.isdir(env.logs_dir))
        self.assertTrue(os.path.isdir(env.pkgs_dir))




# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from gevent import monkey
monkey.patch_all()

import time
import unittest
import mock
import threading
from ava.shell.console import Shell


class ShellTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @mock.patch('ava.shell.base.Supervisor')
    def test_console(self, mock_supervisor):
        mc = mock_supervisor
        ms = mock_supervisor.return_value
        ms.is_server_running.return_value = False
        ms.healthcheck.return_value = True
        shell = Shell()

        self.assertTrue(mc.called)
        shell.is_server_running()
        self.assertTrue(ms.is_server_running.called)
        self.assertFalse(shell.is_server_running())

        t = threading.Thread(target=shell.do_run)
        t.start()
        time.sleep(0.1)
        self.assertTrue(ms.start_server.called)

        ms.is_server_running.return_value = True
        self.assertTrue(shell.is_server_running())
        shell.shell_stopped = True
        t.join()


    @mock.patch('ava.shell.base.Supervisor')
    def test_console_with_bad_health_check(self, mock_supervisor):
        mc = mock_supervisor
        ms = mock_supervisor.return_value
        ms.is_server_running.return_value = False
        ms.health_check.return_value = True
        shell = Shell()

        self.assertTrue(mc.called)
        shell.is_server_running()
        self.assertTrue(ms.is_server_running.called)
        self.assertFalse(shell.is_server_running())

        t = threading.Thread(target=shell.do_run)
        t.start()
        time.sleep(0.1)
        self.assertTrue(ms.start_server.called)

        ms.is_server_running.return_value = True
        self.assertTrue(shell.is_server_running())

        ms.health_check.return_value = False
        t.join()
        self.assertTrue(shell.shell_stopped)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import time

from ava.shell.base import ShellBase


class Shell(ShellBase):
    """
    Shell used in text console.
    """
    def __init__(self):
        super(Shell, self).__init__()

    def run(self):
        while not self.shell_stopped:
            try:
                time.sleep(1)
                super(Shell, self).check_server()
            except KeyboardInterrupt:
                break



# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import

import os
import logging
import AppKit
from AppKit import *

from ava.shell import resource_path
from ava.shell.base import ShellBase, STR_EXIT, STR_STATUS, STR_OPEN_WEBFRONT


logger = logging.getLogger(__name__)


def applicationSupportFolder(self):
    paths = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory,NSUserDomainMask,True)
    basePath = (len(paths) > 0 and paths[0]) or NSTemporaryDirectory()
    fullPath = basePath.stringByAppendingPathComponent_("Ava")
    if not os.path.exists(fullPath):
        os.mkdir(fullPath)
    return fullPath


def pathForFilename(self,filename):
    return self.applicationSupportFolder().stringByAppendingPathComponent_(filename)


class AppDelegate(NSObject):
    def init(self):
        #self = super(AppDelegate, self).init()
        if self is None:
            return None

        # Get objc references to the classes we need.
        self.NSUserNotification = objc.lookUpClass('NSUserNotification')
        self.NSUserNotificationCenter = objc.lookUpClass('NSUserNotificationCenter')

        return self

    def applicationDidFinishLaunching_(self, sender):
    	logger.debug("Application did finish launching.")

        logger.debug("Icon file: %s", resource_path('ava/res/eavatar.png'))
        statusbar = NSStatusBar.systemStatusBar()
        self.statusitem = statusbar.statusItemWithLength_(NSVariableStatusItemLength)
        self.icon = NSImage.alloc().initByReferencingFile_(resource_path('res/eavatar.png'))
        self.icon.setScalesWhenResized_(True)
        self.icon.setSize_((20, 20))
        self.statusitem.setImage_(self.icon)
#    	self.statusitem.setTitle_(u"影化身")
    	self.statusitem.setHighlightMode_(True)
    	self.statusitem.setEnabled_(True)

        #make the menu
        self.menubarMenu = NSMenu.alloc().init()

        self.menuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_(STR_OPEN_WEBFRONT, 'openMainFrame:', '')
        self.menubarMenu.addItem_(self.menuItem)

        self.menuItem = NSMenuItem.separatorItem()
        self.menubarMenu.addItem_(self.menuItem)

        self.quit = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_(STR_EXIT, 'terminate:', '')
        self.menubarMenu.addItem_(self.quit)

        #add menu to statusitem
        self.statusitem.setMenu_(self.menubarMenu)
        self.statusitem.setToolTip_(u'Ava - running')

    def applicationWillTerminate_(self,sender):
        logger.debug("Application will terminate.")

        self.shell.stop_server()
        self.shell.shell_stopped = True

    def alert(self, msg, title="Important Message"):
        self.app.activateIgnoringOtherApps_(True)
        alert = NSAlert.alloc().init()
        alert.setMessageText_(title)
        alert.setInformativeText_(msg)
        alert.setAlertStyle_(NSInformationalAlertStyle)
        alert.runModal()

    def openMainFrame_(self, notification):
        logger.debug('Opening main UI...')

        self.shell.open_main_ui()

    def onQuitApp_(self, notification):
        self.shell.stop_server()

    def hello_(self, notification):
        self.alert("Message from Sam", "Hello There!!!")


class Shell(ShellBase):
    def __init__(self):
        super(Shell, self).__init__()

        self.app = None
        self.delegate = None
        self.mainframe = None

    def doIdle_(self, timer):
        super(Shell, self).check_server()
        if self.shell_stopped and self.app is not None:
            logger.debug("Request to terminate application.")
            self.app.terminate_(1)

    def run(self):
        NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(1, self, 'doIdle:', "", True)
        self.app = NSApplication.sharedApplication()
        self.delegate = AppDelegate.alloc().init()
        self.delegate.shell = self
        self.delegate.app = self.app
        self.app.setDelegate_(self.delegate)
        self.app.activateIgnoringOtherApps_(True)

        self.app.run()

if __name__ == '__main__':
    shell = Shell()
    shell.run()


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import logging

import gevent
import bottle
from ws4py.server.geventserver import WSGIServer
from ava.runtime import settings
from ava.runtime import environ


logger = logging.getLogger(__name__)

_CONF_SECTION = 'webfront'


class ApplicationDispatcher(object):
    """Allows one to mount middlewares or applications in a WSGI application.
    """

    def __init__(self, app, mounts=None):
        self.app = app
        self.mounts = mounts or {}

    def __call__(self, environ, start_response):
        script = environ.get(b'PATH_INFO', b'')
        # logger.debug("initial script: %s", script)
        path_info = ''
        while b'/' in script:
            if script in self.mounts:
                app = self.mounts[script]
                break
            script, last_item = script.rsplit(b'/', 1)
            path_info = b'/%s%s' % (last_item, path_info)
        else:
            # logger.debug("Selected script: %s", script)
            app = self.mounts.get(script, self.app)
        original_script_name = environ.get(b'SCRIPT_NAME', b'')
        environ[b'SCRIPT_NAME'] = original_script_name + script
        environ[b'PATH_INFO'] = path_info
        return app(environ, start_response)

    def mount(self, path, app):
        logger.debug("Mounting app at %s", path)
        self.mounts[path] = app
        return app

    def unmount(self, path):
        logger.debug("Unmounting app at %s", path)
        app = self.mounts.get(path)
        if app is not None:
            del self.mounts[path]
            return app
        return None

# the global web application
dispatcher = ApplicationDispatcher(bottle.app())


class WebfrontEngine(object):
    """
    The client-facing web interface.
    """
    def __init__(self):
        logger.debug("Initializing webfront engine...")
        self._http_listener = None
        self._https_listener = None
        self.listen_port = 5000
        self.listen_addr = '127.0.0.1'
        self.secure_listen_port = 0  # o means not binding
        self.local_base_url = "http://127.0.0.1:%d/" % (self.listen_port,)

    def start(self, ctx=None):
        logger.debug("Starting webfront engine...")

        self.listen_port = settings[_CONF_SECTION]['listen_port']
        self.listen_addr = settings[_CONF_SECTION]['listen_addr']
        self.secure_listen_port = settings[_CONF_SECTION]['secure_listen_port']
        self.local_base_url = "http://127.0.0.1:%d/" % (self.listen_port,)

        logger.debug("Local base URL:%s", self.local_base_url)

        if self.listen_port != 0:
            ctx.add_child_greenlet(gevent.spawn(self._run_http))

        if self.secure_listen_port != 0:
            ctx.add_child_greenlet(gevent.spawn(self._run_https))

        logger.debug("Webfront engine started.")

    def stop(self, ctx=None):
        logger.debug("Webfront engine stopped.")

    def _run_https(self):
        logger.debug("Webfront engine(HTTPS) is running...")

        conf_dir = environ.conf_dir()
        keyfile = os.path.join(conf_dir, 'ava.key')
        certfile = os.path.join(conf_dir, 'ava.crt')

        self._https_listener = WSGIServer(
            (self.listen_addr, self.secure_listen_port),
            dispatcher,
            keyfile=keyfile,
            certfile=certfile)

        logger.debug("Webfront engine(HTTPS) is listening on port: %d",
                     self._https_listener.address[1])

        self._https_listener.serve_forever()

    def _run_http(self):
        logger.debug("Webfront engine(HTTP) is running...")

        self._http_listener = WSGIServer(
            (self.listen_addr, self.listen_port),
            dispatcher)

        logger.debug("Webfront engine(HTTP) is listening on port: %d",
                     self._http_listener.address[1])

        self._http_listener.serve_forever()



# -*- coding: utf-8 -*-
"""
Command for managing local pod directory.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import os
import shutil
import click

from ava.runtime import environ

from .cli import cli


@cli.group()
def pod():
    """ Pod management.
    """
    pass


@pod.command()
@click.argument("folder", type=click.Path(exists=False))
def init(folder):
    """
    Constructs the skeleton of directories if it not there already.
    :return:
    """

    if os.path.exists(folder):
        click.echo("Folder %s is not empty!" % folder, err=True)
        return

    os.makedirs(folder)

    src_dir = environ.pod_dir()
    # copy files from base_dir to user_dir
    subdirs = os.listdir(src_dir)
    for d in subdirs:
        src_path = os.path.join(src_dir, d)
        dst_path = os.path.join(folder, d)
        if os.path.isdir(src_path):
            shutil.copytree(src_path, dst_path)
        else:
            shutil.copy2(src_path, dst_path)


@pod.command()
def open():
    """ Open Pod folder in a file explorer or the like.
    """
    click.launch(environ.pod_dir())


# -*- coding: utf-8 -*-
"""
Data engine implementation based on lighting memory database (http://symas.com/mdb/).
The Lmdb is initialized, the access needs to use its binding API, though.
Extension packages may provide higher-level APIs based on this.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import time
import logging

import lmdb
from ava.runtime import environ


_DATA_FILE_DIR = 'data'

logger = logging.getLogger(__name__)


class DataEngine(object):
    def __init__(self):
        logger.debug("Initializing data engine...")
        self.datapath = None
        self.database = None

    def start(self, ctx=None):
        logger.debug("Starting data engine...")

        self.datapath = os.path.join(environ.pod_dir(), _DATA_FILE_DIR)
        logger.debug("Data path: %s", self.datapath)

        try:
            self.database = lmdb.Environment(self.datapath, max_dbs=32)
        except lmdb.Error:
            logger.exception("Failed to open database.", exc_info=True)
            raise

        logger.debug("Data engine started.")

    def stop(self, ctx=None):
        logger.debug("Stopping data engine...")
        if self.database:
            self.database.close()

        logger.debug("Data engine stopped.")




# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function)

import os
import unittest
from yaml import load, dump
from ava.util import base_path

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


class YamlConfigTest(unittest.TestCase):
    def setUp(self):
        self.conf_file = os.path.join(base_path(), 'tests', 'data', 'test.yml')

    def test_can_load_yaml_file(self):
        fin = file(self.conf_file)
        data = load(fin, Loader=Loader)
        # print(data)
        self.assertTrue(isinstance(data, dict))

    def test_can_dump_yaml_file(self):
        data = {
            'a': "a",
            'b': "b",
            'sample': {
                'k1': 'v1'
            }
        }
        self.assertTrue(isinstance(data, dict))
        output = dump(data, Dumper=Dumper)

        self.assertTrue("sample" in output)


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import unittest

from ava.spi.context import Context


class Receiver(object):
    def __init__(self):
        self.called = False

    def __call__(self, *args, **kwargs):
        self.called = True


class CoreContextTest(unittest.TestCase):

    def setUp(self):
        self.context = Context(None)

    def test_send_signal(self):
        SIGNAL = 'my-first-signal'

        receiver = Receiver()
        self.context.connect(receiver, signal=SIGNAL)
        self.context.send(signal=SIGNAL)
        self.assertTrue(receiver.called)

    def test_connect_and_then_disconnect(self):
        SIGNAL = 'my-second-signal'
        receiver = Receiver()
        self.context.connect(receiver)
        self.context.send(signal=SIGNAL)
        self.assertTrue(receiver.called)
        receiver.called = False
        self.context.disconnect(receiver)
        self.context.send(signal=SIGNAL)
        self.assertFalse(receiver.called)


# -*- coding: utf-8 -*-

from __future__ import print_function

import os
import unittest
import lmdb
import tempfile
import shutil


class TestLMDB(unittest.TestCase):

    def setUp(self):
        self.path = tempfile.mkdtemp()
        self.lmdb = lmdb.Environment(self.path, max_dbs=4)

    def tearDown(self):
        self.lmdb.close()
        shutil.rmtree(self.path)

    def test_info(self):
        print(self.lmdb.info())
        print("Max key size: ", self.lmdb.max_key_size())

    def test_crud(self):
        with self.lmdb.begin(write=True) as txn:
            txn.put("k1", "value1")

            v1 = txn.get("k1")
            self.assertEqual(v1, "value1")

            txn.put("k1", "value2")
            v2 = txn.get("k1")
            self.assertEqual(v2, "value2")

            txn.delete("k1")

            v3 = txn.get("k1")
            self.assertTrue(v3 is None)

    def test_subdb(self):
        db = self.lmdb.open_db("test", create=True)

        self.assertIsNotNone(db, "db should not be none.")

        with self.lmdb.begin(write=False) as txn:
            cur = txn.cursor()
            for k, v in iter(cur):
                print("Found db: ", k)





# -*- coding: utf-8 -*-
"""
Extension engine is responsible for adding packages(in egg format) to the path,
and then starting or stopping extensions.

Some packages provides no extension but acts as libraries for other packages.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import pkg_resources

from ava.runtime.package import PackageManager


logger = logging.getLogger(__name__)


class Extension(object):
    def __init__(self, name, entry_point):
        self.name = name
        self.entry_point = entry_point
        self.cls = None
        self.obj = None


class ExtensionManager(object):
    def __init__(self, namespace="ava.extension"):
        self.namespace = namespace
        self.extensions = []

    def load_extensions(self, invoke_on_load=True):
        for it in pkg_resources.working_set.iter_entry_points(self.namespace, name=None):
            logger.debug("Loading extension: %s at module: %s", it.name, it.module_name)
            logger.debug("")
            extension = it.load()
            ext = Extension(it.name, it)
            ext.cls = it.load()

            if invoke_on_load:
                ext.obj = ext.cls()
            self.extensions.append(ext)

    def start_extensions(self, context=None):
        for ext in self.extensions:
            startfun = getattr(ext.obj, "start", None)
            if startfun is not None and callable(startfun):
                    startfun(context)

    def stop_extensions(self, context=None):
        for ext in self.extensions:
            stopfun = getattr(ext.obj, "stop", None)
            if stopfun is not None and callable(stopfun):
                try:
                    stopfun(context)
                except Exception:
                    pass
                    #IGNORED.


class ExtensionEngine(object):
    """
    Responsible for managing extension packages.
    """
    def __init__(self):
        self._extension_mgr = ExtensionManager()
        self._package_mgr = PackageManager()

    def start(self, ctx):
        logger.debug("Starting extension engine...")
        self._package_mgr.find_packages()
        self._extension_mgr.load_extensions()
        self._extension_mgr.start_extensions(ctx)
        logger.debug("Extension engine started.")

    def stop(self, ctx):
        logger.debug("Stopping extension engine...")
        self._extension_mgr.stop_extensions(ctx)
        logger.debug("Extension engine stopped.")


# -*- coding: utf-8 -*-
"""
For generating monotonically increasing UUIDs.
"""
from __future__ import print_function, division, absolute_import

import uuid
import datetime
import calendar
import threading
import random


def utcnow():
    return TimeUUID.with_utcnow()


def utctime():
    """
    Generate a timestamp from the current time in UTC. Returns exactly
    the same thing as :py:func`time.time` but with a UTC offset instead
    of a local one

    :returns: a timestamp
    :rtype: float
    """
    d = datetime.datetime.utcnow()
    return mkutime(d)


def mkutime(d):
    """
    Generate a UTC timestamp from the datetime object.

    :param d: a :py:class:`datetime` object
    :type d: datetime
    """
    return float(calendar.timegm(d.timetuple())) + float(d.microsecond) / 1e6


class IncreasingMicrosecondClock(object):
    """
    A clock that returns a new timestamp value unique across all threads,
    achieved by locking a mutex each call and checking the previous value.
    """
    def __init__(self, timestamp_factory=utctime, mutex=threading.Lock):
        self.timestamp_factory = timestamp_factory
        self.mutex = mutex()
        self.time = timestamp_factory()

    def __call__(self):
        with self.mutex:
            new_time = self.timestamp_factory()
            if new_time > self.time:
                self.time = new_time
            else:
                self.time += .000001
            return self.time


class TimeUUID(uuid.UUID):
    """
    A class that makes dealing with time and V1 UUIDs much easier. Offers
    accurate comparison (compares accurately with other TimeUUIDs and UUIDv1),
    a way to get TimeUUIDs with UTC-timestamps (instead of the default
    localtime), an easy way to get the UTC datetime object from it, and
    default encoding and decoding methods.
    """
    timestamp_factory = IncreasingMicrosecondClock()

    def __repr__(self):
        return u"TimeUUID('{}')".format(super(TimeUUID, self).__str__())

    def __cmp__(self, other):
        # if other isn't a UUID, do whatever it would do WWUUIDD?
        if not isinstance(other, uuid.UUID):
            return super(TimeUUID, self).__cmp__(other)
        if self.version == 1 and other.version == 1:
            if self.time == other.time:
                return super(TimeUUID, self).__cmp__(other)
            else:
                return cmp(self.time, other.time)
        else:
            return super(TimeUUID, self).__cmp__(other)

    def get_datetime(self):
        return datetime.datetime.utcfromtimestamp(self.get_timestamp())

    def get_timestamp(self):
        return (self.get_time() - 0x01b21dd213814000L) / 1e7

    @classmethod
    def upgrade(cls, other):
        if other is None:
            return other
        elif isinstance(other, cls):
            return other
        elif isinstance(other, uuid.UUID):
            return cls(hex=other.hex)
        raise TypeError("Cannot upgrade %s because it is not a uuid.UUID "
                        "(it is a '%s')" % (other, other.__class__.__name__))

    @classmethod
    def with_utc(cls, d, randomize=True, lowest_val=False):
        """
        Create a TimeUUID with any datetime in UTC. Only use this if you are sure
        you are creating your datetime objects properly and can predict how
        the :py:class:`TimeUUID` will sort (if that is at all necessary).

        See :py:meth:`with_timestamp` for a description of how `randomize` and
        `lowest_val` arguments work.

        :param d: A datetime object
        :type d: datetime

        :returns: A TimeUUID object
        :type d: TimeUUID
        """
        return cls.with_timestamp(mkutime(d), randomize=randomize,
                                  lowest_val=lowest_val)

    @classmethod
    def with_utcnow(cls, randomize=True):
        """
        Create a TimeUUID with the current datetime in UTC. Every TimeUUID
        generated this way, on this machine, will supersede the TimeUUID generated
        before it, by use of a mutex.

        :param randomize: Whether or not to randomize the other bits in the UUID.
        Defaults to `True`
        :type randomize: bool

        :returns: A TimeUUID object
        :rtype: TimeUUID
        """
        return cls.with_timestamp(cls.timestamp_factory(), randomize=randomize)

    @classmethod
    def convert(cls, value, randomize=True, lowest_val=False):
        """
        Try to convert a number of different types to a TimeUUID. Works for
        datetimes, ints (which it treats as timestamps) and UUIDs.

        See :py:meth:`with_timestamp` for descriptions of the `randomize` and
        `lowest_val` arguments.

        :param value: An instance of UUID, TimeUUID, datetime or timestamp
            int or float

        :returns: A TimeUUID object
        :rtype: TimeUUID
        """
        if isinstance(value, cls):
            return value
        if isinstance(value, uuid.UUID):
            return cls.upgrade(value)
        if isinstance(value, datetime.datetime):
            return cls.with_timestamp(mkutime(value), randomize=randomize,
                lowest_val=lowest_val)
        if isinstance(value, (int, long, float)):
            return cls.with_timestamp(value, randomize=randomize,
                lowest_val=lowest_val)
        raise ValueError("Sorry, I don't know how to convert {} to a "
                         "TimeUUID".format(value))

    @classmethod
    def with_timestamp(cls, timestamp, randomize=True, lowest_val=False):
        """
        Create a TimeUUID with any timestamp. Be sure that the timestamp is
        created with the correct offset. If you are using :py:func:`time.time`
        be sure to know that in python, it may be offset by your machine's
        time zone!

        :param timestamp: A timestamp, like that returned by :py:func:`utctime`
            or py:func:`time.time`.
        :type timestamp: float
        :param randomize: will create a UUID with randomish bits. This overrides
            it's counterpart: `lowest_val` If randomize is False and `lowest_val`
            is False, it will generate a UUID with the highest possible clock seq
            for the same timestamp, otherwise generating a UUID with the *lowest*
            possible clock seq for the same timestamp. Defaults to `True`
        :type randomize: bool
        :param lowest_val: will create a UUID with the lowest possible clock seq if
            randomize is False if set to True, otherwise generating the highest
            possible. Defaults to `False`
        :type lowest_val: bool

        :returns: A TimeUUID object
        :rtype: TimeUUID
        """
        ns = timestamp * 1e9
        ts = int(ns // 100) + 0x01b21dd213814000L
        time_low = ts & 0xffffffffL
        time_mid = (ts >> 32L) & 0xffffL
        time_hi_version = (ts >> 48L) & 0x0fffL
        if randomize:
            cs = random.randrange(1<<14L)
            clock_seq_low = cs & 0xffL
            clock_seq_hi_variant = (cs >> 8L) & 0x3fL
            node = uuid.getnode()
        else:
            if lowest_val: # uuid with lowest possible clock value
                clock_seq_low = 0 & 0xffL
                clock_seq_hi_variant = 0 & 0x3fL
                node = 0 & 0xffffffffffffL # 48 bits
            else: # UUID with highest possible clock value
                clock_seq_low = 0xffL
                clock_seq_hi_variant = 0x3fL
                node = 0xffffffffffffL # 48 bits
        return cls(
            fields=(time_low, time_mid, time_hi_version,
                    clock_seq_hi_variant, clock_seq_low, node),
            version=1
        )


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys


# helper function for constructing paths to resource files.
def resource_path(relative):
    if hasattr(sys, "_MEIPASS"):
        return os.path.join(sys._MEIPASS, relative)

    res_path = os.environ.get("RESOURCEPATH", None)
    if res_path is None:
        abspath = os.path.abspath(os.path.join(__file__, "..", ".."))
        abspath = os.path.dirname(abspath)
        #print("abspath: ", abspath)
        return os.path.join(abspath, relative)
    else:
        return os.path.join(res_path, relative)






# -*- coding: utf-8 -*-

import unittest

from ava.runtime import settings


class RuntimeConfigTests(unittest.TestCase):

    def test_should_have_dir_settings(self):
        self.assertIsNotNone(settings.get('pkgs_dir'))
        self.assertIsNotNone(settings.get('conf_dir'))
        self.assertIsNotNone(settings.get('data_dir'))
        self.assertIsNotNone(settings.get('logs_dir'))

    def test_should_have_settings_for_webfront(self):
        self.assertIsNotNone(settings['webfront']['listen_port'])

    def test_should_have_logging_settings(self):
        handlers = settings['logging']['handlers']
        self.assertIsNotNone(handlers)
        log_file = handlers['file_handler']['filename']
        #print(log_file)
        self.assertIsNotNone(log_file)





# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
Imports utility functions from Bottle for exposing web resource.
"""
import os
from ava.runtime import environ
import logging
import six
import binascii
from bottle import request, hook, HTTPError
from ava.util import token
from ava.util import crypto
from ava.runtime.config import settings

from bottle import route, get, post, delete, put, request, response
from bottle import static_file as _static_file
from ava.core.webfront import dispatcher

logger = logging.getLogger(__name__)


static_folder = os.path.join(environ.pod_dir(), 'webroot')


def _raise_unauthorized(desc=b'Authentication required.'):
    raise HTTPError(401, desc)


def check_authentication():
    logger.debug("Checking authentication...")

    http_auth = request.get_header('Authorization')

    logger.debug("Authentication header: %s", http_auth)

    if http_auth is None:
        msg = "No authentication header value is given."
        _raise_unauthorized(desc=msg)

    if isinstance(http_auth, six.string_types):
            http_auth = http_auth.encode('ascii')

    auth_type = ""
    user_and_key = ""

    try:
        auth_type, user_and_key = http_auth.split(six.b(' '), 1)
    except ValueError as err:
        msg = ("Basic authorize header value not properly formed. "
               "Supplied header {0}. Got error: {1}")
        msg = msg.format(http_auth, str(err))
        logger.debug(msg)
        _raise_unauthorized(desc=msg)

    auth_type = auth_type.lower()

    if auth_type == six.b('basic'):
        request.client_xid = 'tester'
        return

    if auth_type == six.b('basic'):
        logging.debug("Authenticating clint using Basic mechanism...")
        try:
            user_and_key = user_and_key.strip()
            user_and_key = binascii.a2b_base64(user_and_key)
            user_id, key = user_and_key.split(six.b(':'), 1)

            logger.debug("user_id: %s", user_id)

            if not crypto.validate_xid(user_id):
                _raise_unauthorized()

            xid = crypto.secret_key_to_xid(key)
            if xid != user_id:
                _raise_unauthorized()

            request.client_xid = user_id
            logger.debug("Client authenticated: %s", user_id)
            return
        except (binascii.Error, ValueError) as err:
            msg = ("Unable to determine user and pass/key encoding. "
                   "Got error: {0}").format(str(err))
            logger.debug(msg)
            _raise_unauthorized(desc=msg)

    elif auth_type == six.b("bearer"):
        logger.debug("Authenticating client using Bearer mechanism...")
        try:
            tok = token.decode(user_and_key, settings.NETWORK_SECRET, verify=True)
        except token.DecodeError:
            logger.debug("Token decode error", exc_info=True)
            tok = {}
            _raise_unauthorized()

        client_xid = tok.get("sub")
        if not client_xid:
            _raise_unauthorized(desc="Subject not specified in token.")

        if not crypto.validate_xid(client_xid):
            _raise_unauthorized()
        request.client_xid = client_xid
    else:
        _raise_unauthorized("Unsupported authentication method: %s" % auth_type)


def static_file(filepath, root=static_folder, mimetype='auto', download=False, charset='utf-8'):
    return _static_file(filepath, root=root, mimetype=mimetype, download=download, charset=charset)


def swap_root_app(wsgiapp):
    """ Swap the root WSGI application.

    :param wsgiapp:
    :return: the previous WSGI application.
    """
    from ava.core.webfront import dispatcher
    old_app = dispatcher.app
    dispatcher.app = wsgiapp

    return old_app

__all__ = [route, get, post, delete, put, request, response,
           static_file, static_folder, dispatcher, check_authentication]




# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from yaml import load as _load, dump as _dump

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


def load(stream):
    return _load(stream, Loader=Loader)


def dump(data, stream, **options):
    return _dump(data,
                 stream,
                 Dumper=Dumper,
                 default_flow_style=False,
                 indent=4, width=80,
                 **options)


# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import

import os

import win32api
import win32con
import win32gui_struct


try:
    import winxpgui as win32gui
except ImportError:
    import win32gui

import itertools
import glob

from ava.shell import resource_path
from ava.shell.base import ShellBase, STR_EXIT, STR_OPEN_WEBFRONT, STR_STATUS


class MainFrame(object):
    def __init__(self, message_map):
        self.window_class_name = "EAvatarWnd"
        self.hinst = None
        self.class_atom = self.register_wnd_class(message_map)
        self.hwnd = self.create_window()

    def register_wnd_class(self, message_map):
        # Register the Window class.
        window_class = win32gui.WNDCLASS()
        self.hinst = window_class.hInstance = win32gui.GetModuleHandle(None)
        window_class.lpszClassName = self.window_class_name
        window_class.style = win32con.CS_VREDRAW | win32con.CS_HREDRAW
        window_class.hCursor = win32gui.LoadCursor(0, win32con.IDC_ARROW)
        window_class.hbrBackground = win32con.COLOR_WINDOW
        window_class.lpfnWndProc = message_map  # could also specify a wndproc.
        return win32gui.RegisterClass(window_class)

    def create_window(self):
        # Create the Window.
        style = win32con.WS_OVERLAPPED | win32con.WS_SYSMENU
        hwnd = win32gui.CreateWindow(self.class_atom,
                                          self.window_class_name,
                                          style,
                                          0,
                                          0,
                                          310,
                                          250,
                                          0,
                                          0,
                                          self.hinst,
                                          None)
        win32gui.UpdateWindow(hwnd)
        return hwnd

    def show(self):
        win32gui.ShowWindow(self.hwnd, win32con.SW_NORMAL)

    def close(self):
        win32gui.PostQuitMessage(0)

_QUIT = 'QUIT'

_FIRST_ID = 1023
_ID_OPEN_FRAME = 1024
_ID_QUIT = 1026


class StatusIcon(object):
    def __init__(self, s):
        self.shell = s

        self.icons = itertools.cycle(glob.glob(resource_path('res/*.ico')))
        self.hover_text =STR_STATUS

        self.menu_options = ((STR_OPEN_WEBFRONT, None, None, _ID_OPEN_FRAME),
                             ("-", None, None, 1025),
                             (STR_EXIT, None, None, _ID_QUIT),)

        self.icon = self.icons.next()

        self.default_menu_index = 0

        self.notify_id = None
        self.hicon = None
        self.refresh_icon()

    def notify(self, title, message):
        balloon_id = (self.shell.main_frame.hwnd,
                          0,
                          win32gui.NIF_INFO,
                          win32con.WM_USER+20,
                          self.hicon,
                          self.hover_text,
                          title,
                          200,
                          message)
        win32gui.Shell_NotifyIcon(win32gui.NIM_MODIFY, balloon_id)

    def refresh_icon(self):
        # Try and find a custom icon
        hinst = win32gui.GetModuleHandle(None)
        if os.path.isfile(self.icon):
            icon_flags = win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE
            self.hicon = win32gui.LoadImage(hinst,
                                       self.icon,
                                       win32con.IMAGE_ICON,
                                       0,
                                       0,
                                       icon_flags)
        else:
            print("Can't find icon file - using default.")
            self.hicon = win32gui.LoadIcon(0, win32con.IDI_APPLICATION)

        if self.notify_id:
            message = win32gui.NIM_MODIFY
        else:
            message = win32gui.NIM_ADD

        self.notify_id = (self.shell.main_frame.hwnd,
                          0,
                          win32gui.NIF_ICON | win32gui.NIF_MESSAGE | win32gui.NIF_TIP,
                          win32con.WM_USER+20,
                          self.hicon,
                          self.hover_text)
        win32gui.Shell_NotifyIcon(message, self.notify_id)

    def show_menu(self):
        menu = win32gui.CreatePopupMenu()
        self.create_menu(menu, self.menu_options)

        pos = win32gui.GetCursorPos()
        # See http://msdn.microsoft.com/library/default.asp?url=/library/en-us/winui/menus_0hdi.asp
        win32gui.SetForegroundWindow(self.shell.main_frame.hwnd)
        win32gui.TrackPopupMenu(menu,
                                win32con.TPM_LEFTALIGN,
                                pos[0],
                                pos[1],
                                0,
                                self.shell.main_frame.hwnd,
                                None)

    def create_menu(self, menu, menu_options):
        for option_text, option_icon, option_action, option_id in menu_options[::-1]:
            if option_icon:
                option_icon = self.prep_menu_icon(option_icon)

            if option_text == "-":
                win32gui.InsertMenu(menu,  0, win32con.MF_BYPOSITION, win32con.MF_SEPARATOR, None)
            else:
                item, extras = win32gui_struct.PackMENUITEMINFO(text=option_text,
                                                                hbmpItem=option_icon,
                                                                wID=option_id)
                win32gui.InsertMenuItem(menu, 0, 1, item)

    def prep_menu_icon(self, icon):
        # First load the icon.
        ico_x = win32api.GetSystemMetrics(win32con.SM_CXSMICON)
        ico_y = win32api.GetSystemMetrics(win32con.SM_CYSMICON)
        hicon = win32gui.LoadImage(0, icon, win32con.IMAGE_ICON, ico_x, ico_y, win32con.LR_LOADFROMFILE)

        hdcBitmap = win32gui.CreateCompatibleDC(0)
        hdcScreen = win32gui.GetDC(0)
        hbm = win32gui.CreateCompatibleBitmap(hdcScreen, ico_x, ico_y)
        hbmOld = win32gui.SelectObject(hdcBitmap, hbm)
        # Fill the background.
        brush = win32gui.GetSysColorBrush(win32con.COLOR_MENU)
        win32gui.FillRect(hdcBitmap, (0, 0, 16, 16), brush)
        # unclear if brush needs to be feed.  Best clue I can find is:
        # "GetSysColorBrush returns a cached brush instead of allocating a new
        # one." - implies no DeleteObject
        # draw the icon
        win32gui.DrawIconEx(hdcBitmap, 0, 0, hicon, ico_x, ico_y, 0, 0, win32con.DI_NORMAL)
        win32gui.SelectObject(hdcBitmap, hbmOld)
        win32gui.DeleteDC(hdcBitmap)

        return hbm

    def switch_icon(self):
        self.icon = self.icons.next()
        self.refresh_icon()


class Shell(ShellBase):
    def __init__(self):
        super(Shell, self).__init__()

        msg_taskbar_restart = win32gui.RegisterWindowMessage("TaskbarCreated")
        self.message_map = {msg_taskbar_restart: self.OnRestart,
                       win32con.WM_DESTROY: self.OnDestroy,
                       win32con.WM_COMMAND: self.OnCommand,
                       win32con.WM_USER+20: self.OnTaskbarNotify,}

        self.main_frame = MainFrame(self.message_map)
        self.status_icon = StatusIcon(self)

    def run(self):
        #while not win32gui.PumpWaitingMessages():
        #    time.sleep(0.1)
        win32gui.PumpMessages()

    def OnCommand(self, hwnd, msg, wparam, lparam):
        id = win32gui.LOWORD(wparam)
        self.execute_menu_option(id)

    def OnRestart(self, hwnd, msg, wparam, lparam):
        self.status_icon.refresh_icon()

    def OnDestroy(self, hwnd, msg, wparam, lparam):
        self.stop_server()

        nid = (self.main_frame.hwnd, 0)
        win32gui.Shell_NotifyIcon(win32gui.NIM_DELETE, nid)
        win32gui.PostQuitMessage(0)  # Terminate the app.

    def OnTaskbarNotify(self, hwnd, msg, wparam, lparam):
        if lparam == win32con.WM_LBUTTONDBLCLK:
            self.status_icon.execute_menu_option(self.FIRST_ID)
        elif lparam == win32con.WM_RBUTTONUP:
            self.status_icon.show_menu()
        elif lparam == win32con.WM_LBUTTONUP:
            pass

        return True

    def execute_menu_option(self, id):
        if id == _ID_QUIT:
            win32gui.DestroyWindow(self.main_frame.hwnd)
        elif id == _ID_OPEN_FRAME:
            self.open_main_ui()


if __name__ == '__main__':
    shell = Shell()
    shell.run()


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import unittest
import re

from ava.util import resource_path, time_uuid
from ava.util import webutils


class UtilTest(unittest.TestCase):

    def test_resource_path(self):
        rv = resource_path("path1")
        print(rv)
        self.assertTrue(rv.endswith("path1"))

    def test_time_uuid(self):
        prev = time_uuid.utcnow()
        for i in xrange(1000):
            now = time_uuid.utcnow()
            #print(now.hex)
            self.assertTrue(now > prev)
            prev = now


    def test_parse_authorization_header(self):
        result = webutils.parse_authorization_header('EAvatar key="1234",realm="http://eavatar.me", token="abcd"')
        self.assertEqual("eavatar", result['scheme'])
        self.assertEqual("http://eavatar.me", result["realm"])
        self.assertEqual("1234", result["key"])
        self.assertEqual("abcd", result["token"])

    def test_store_url_without_path(self):
        store_without_path = '/a/QPndzFJTmycdfg5jxcSghX2scJnc3TNqVEfYtVTA5JVYiPQY/store'
        store_with_path = '/a/QPndzFJTmycdfg5jxcSghX2scJnc3TNqVEfYtVTA5JVYiPQY/store/this/is/a/path?q=test'
        pattern = '^\/a/([a-zA-Z0-9]+)/store(/[^\?]*)?'
        pat = re.compile(pattern)
        matches = pat.match(store_without_path)
        self.assertEqual('/a/QPndzFJTmycdfg5jxcSghX2scJnc3TNqVEfYtVTA5JVYiPQY/store', matches.group(0))
        self.assertEqual('QPndzFJTmycdfg5jxcSghX2scJnc3TNqVEfYtVTA5JVYiPQY', matches.group(1))
        self.assertIsNone(matches.group(2))

    def test_store_url_with_path(self):
        store_with_path = '/a/QPndzFJTmycdfg5jxcSghX2scJnc3TNqVEfYtVTA5JVYiPQY/store/this/is/a/path?q=test'
        pattern = '^\/a/(?P<aid>[a-zA-Z0-9]+)/store(?P<path>/[^\?]*)?'
        pat = re.compile(pattern)
        matches = pat.match(store_with_path)
        self.assertEqual('/a/QPndzFJTmycdfg5jxcSghX2scJnc3TNqVEfYtVTA5JVYiPQY/store/this/is/a/path', matches.group(0))
        self.assertEqual('QPndzFJTmycdfg5jxcSghX2scJnc3TNqVEfYtVTA5JVYiPQY', matches.group('aid'))
        self.assertEqual('/this/is/a/path', matches.group('path'))


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


import click

from .cli import cli
from ava.spi.defines import VERSION_STRING


@cli.command()
def version():
    """ Show version information.

    :return:
    """
    click.echo("Ava Ver. %s" % VERSION_STRING)



# -*- coding: utf-8 -*-
"""
Misc. utility classes/functions.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys


# helper function for constructing paths to resource files.
def resource_path(relative):
    if hasattr(sys, "_MEIPASS"):
        return os.path.join(sys._MEIPASS, relative)

    abspath = os.path.abspath(os.path.join(__file__, "..", ".."))
    abspath = os.path.dirname(abspath)
    #print("abspath: ", abspath)
    return os.path.join(abspath, relative)


# helper function for constructing paths to resource files.
def base_path():
    if hasattr(sys, "_MEIPASS"):
        return sys._MEIPASS
    else:
        # assumes this file is located at src/eavatar/util/__init__.py
        abspath = os.path.abspath(os.path.join(__file__, "..", ".."))
        abspath = os.path.dirname(abspath)
        return abspath


def is_frozen():
    """ Checks if in frozen mode.
    :return:
    """
    return hasattr(sys, "frozen")

if __name__=='__main__':
    print(resource_path('test'))



# -*- coding: utf-8 -*-
"""
Exceptions raised in eavtar.kits package.
"""
from __future__ import absolute_import, division, print_function, unicode_literals


class AvaError(Exception):
    """
    Raised when error is framework-related but no specific error subclass exists.
    """
    def __init__(self, *args, **kwargs):
        super(AvaError, self).__init__(args, kwargs)


class DataError(AvaError):
    """
    Generic error related to database operations.
    """
    pass


class TaskAlreadyRegistered(AvaError):
    """ Raised to indicate a task is being registered with the same key as a
    existing one.
    """
    def __init__(self, task_key):
        super(TaskAlreadyRegistered, self).__init__()
        self.task_key = task_key

    def __str__(self):
        return "Task %s already registered." % self.task_key


class TaskNotRegistered(AvaError):
    """
    Raised when a task is not registered but is requested to schedule.
    """
    def __init__(self, task_key):
        super(TaskNotRegistered, self).__init__()
        self.task_key = task_key

    def __str__(self):
        return "Task %s not registered." % self.task_key




# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import click
import pip
from .cli import cli
from ava.runtime import environ


@cli.group()
def pkg():
    """ package management.
    """
    pass


@pkg.command()
@click.argument("query", type=click.STRING)
def search(query):
    pip.main(['search', query])

@pkg.command()
@click.argument("requirement")
def install(requirement):
    pip.main(['install', '--target=%s' % environ.pkgs_dir(), requirement])


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


from .cli import cli
from .agent import run
from .info import version
from .pod import init
from .key import validate, generate



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import logging
import gevent
import unittest
from ava.core.task import TaskEngine
from ava.spi.context import Context
from ava.spi.task import task

counter = 0



class TaskEngineTests(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)
        self.engine = TaskEngine()
        self.ctx = Context(None)
        self.engine.start(self.ctx)
        self.logger = logging.getLogger(__name__)

    def tearDown(self):
        self.engine.stop(self.ctx)

    def test_register_and_unregister_task(self):

        def mock_task1():
            return True

        t1 = self.engine.register(mock_task1)
        task_key = __name__ + '.mock_task1'
        self.assertEqual(t1.key, task_key)

        self.engine.unregister(task_key)

        self.assertIsNone(self.engine.get_task(task_key))

    def test_once_schedule(self):
        def mock_task2():
            self.logger.debug("mock_task2() run.")
            return True

        t1 = self.engine.register(mock_task2)
        task_key = __name__ + '.mock_task2'
        schedule1 = t1.run_once()
        schedule1.join()
        self.assertTrue(schedule1.result)

    def test_periodic_schedule(self):
        global counter
        counter = 0

        def mock_task3():
            global counter
            counter += 1
            return counter

        t1 = self.engine.register(mock_task3)

        sched = t1.run_periodic(0.1)
        gevent.sleep(1)
        self.assertTrue(sched.result > 5)

    def test_cancel_schedule(self):
        def mock_once_task():
            return True

        t1 = self.engine.register(mock_once_task)
        schedule1 = t1.run_once(delayed_secs=10)
        got_sched = self.engine.get_schedule(schedule1.id)
        self.assertIsNotNone(got_sched)

        self.engine.cancel(schedule1)
        got_sched = self.engine.get_schedule(schedule1.id)
        self.assertIsNone(got_sched)



# -*- coding: utf-8 -*-
"""
For serving static content in webroot folder.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from ava.spi.webfront import route, static_file


@route('/')
@route('/<filepath:path>')
def serve_static(filepath='index.html'):
    return static_file(filepath)




# -*- coding: utf-8 -*-

"""
Configuration file reading/writing.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import codecs
import logging
import logging.config
import os.path
from bottle import template

from yaml import load, dump

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from ava.runtime import environ


AGENT_CONF = os.path.join(environ.conf_dir(), u'ava.yml')

# The default configuration file is located at the base directory.

settings = dict(base_dir=environ.base_dir(),
                conf_dir=environ.conf_dir(),
                data_dir=environ.data_dir(),
                pkgs_dir=environ.pkgs_dir(),
                logs_dir=environ.logs_dir(),
                mods_dir=environ.mods_dir(),
                )


def load_conf(conf_file):
    if not os.path.exists(conf_file):
        return {}

    data = codecs.open(conf_file, 'rb', encoding='utf-8').read()
    if len(data.strip()) == 0:
        return {}

    data = template(data, **settings)
    return load(data, Loader=Loader)


def save_conf(conf_file, content):
    out = codecs.open(conf_file, 'wb', encoding='utf-8')
    out.write(dump(content, Dumper=Dumper, default_flow_style=False,
                   indent=4, width=80))

settings.update(load_conf(AGENT_CONF))

# configure logging
logging.config.dictConfig(settings['logging'])



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import glob
import logging
import click
from importlib import import_module
from ava.runtime import environ

_MODULES_DIR = os.path.join('mods', 'commands')

# the package name for modules.
_MODULE_PKG = 'mods.commands.'

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

logger = logging.getLogger(__name__)

_modules_path = os.path.join(environ.pod_dir(), _MODULES_DIR)
_modules_path = os.path.abspath(_modules_path)
_modules = {}


def _scan_modules():
    pattern = os.path.join(_modules_path, '[a-zA-Z][a-zA-Z0-9_]*.py')
    return glob.glob(pattern)


def load_commands():
    sys.path.append(environ.pod_dir())
    # logger.debug("Command module directory: %s", _modules_path)

    module_files = _scan_modules()

    # logger.debug("Found %d command module(s)" % len(module_files))

    for s in module_files:
        name = os.path.basename(s)
        if '__init__.py' == name:
            continue

        # gets the basename without extension part.
        name = os.path.splitext(name)[0]
        try:
            # logger.debug("Loading command module: %s", name)
            mod = import_module(_MODULE_PKG + name)
            _modules[name] = mod
        except ImportError:
            logger.error("Failed to import command module: %s",
                         name,
                         exc_info=True)


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True)
@click.pass_context
def cli(ctx, verbose):
    """ The command-line interface for Ava.
    """
    ctx.obj = dict(verbosity=verbose)

    log_level = logging.WARNING

    if verbose > 1:
        log_level = logging.DEBUG
    elif verbose == 1:
        log_level = logging.INFO

    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s', level=log_level, disable_existing_loggers=False)

    app_dir = click.get_app_dir('.ava')
    if not os.path.exists(app_dir):
        os.makedirs(app_dir, 0700)
    if verbose > 2:
        click.echo("Location for configuration files: %s" % app_dir)

    ctx.obj['app_dir'] = app_dir

    if ctx.invoked_subcommand is None:
        from .agent import run
        ctx.invoke(run)


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import logging
import hashlib
try:
    from urllib2 import parse_http_list as _parse_list_header
except ImportError: # pragma: no cover
    from urllib.request import parse_http_list as _parse_list_header
from ava.util import resource_path


static_folder = resource_path('static')

logger = logging.getLogger(__name__)

_ext_to_media_type = {
    '.jpg': 'image/jpeg',
    '.png': 'image/png',
    '.ico': 'image/vnd.microsoft.icon',
    '.svg': 'image/svg+xml',
    '.txt': 'text/plain',
    '.html': 'text/html; charset=utf-8',
    '.css': 'text/css',
    '.js': 'application/javascript',
    '.json': 'application/json',
}

_default_media_type = 'application/octet-stream'


def calc_etag(content):
    md5 = hashlib.md5()
    md5.update(content)
    return md5.hexdigest()


def guess_media_type(ext):
    t = _ext_to_media_type.get(ext)
    if t is None:
        return _default_media_type
    else:
        return t


def unquote_header_value(value, is_filename=False):
    r"""Unquotes a header value.  (Reversal of :func:`quote_header_value`).
    This does not use the real unquoting but what browsers are actually
    using for quoting.

    .. versionadded:: 0.5

    :param value: the header value to unquote.
    """
    if value and value[0] == value[-1] == '"':
        # this is not the real unquoting, but fixing this so that the
        # RFC is met will result in bugs with internet explorer and
        # probably some other browsers as well.  IE for example is
        # uploading files with "C:\foo\bar.txt" as filename
        value = value[1:-1]

        # if this is a filename and the starting characters look like
        # a UNC path, then just return the value without quotes.  Using the
        # replace sequence below on a UNC path has the effect of turning
        # the leading double slash into a single slash and then
        # _fix_ie_filename() doesn't work correctly.  See #458.
        if not is_filename or value[:2] != '\\\\':
            return value.replace('\\\\', '\\').replace('\\"', '"')
    return value


def parse_list_header(value):
    """Parse lists as described by RFC 2068 Section 2.

    In particular, parse comma-separated lists where the elements of
    the list may include quoted-strings.  A quoted-string could
    contain a comma.  A non-quoted string could have quotes in the
    middle.  Quotes are removed automatically after parsing.

    It basically works like :func:`parse_set_header` just that items
    may appear multiple times and case sensitivity is preserved.

    The return value is a standard :class:`list`:

    >>> parse_list_header('token, "quoted value"')
    ['token', 'quoted value']

    To create a header from the :class:`list` again, use the
    :func:`dump_header` function.

    :param value: a string with a list header.
    :return: :class:`list`
    """
    result = []
    for item in _parse_list_header(value):
        if item[:1] == item[-1:] == '"':
            item = unquote_header_value(item[1:-1])
        result.append(item)
    return result


def parse_dict_header(value):
    """Parse lists of key, value pairs as described by RFC 2068 Section 2 and
    convert them into a python dict (or any other mapping object created from
    the type with a dict like interface provided by the `cls` argugment):

    :param value: a string with a dict header.
    :param cls: callable to use for storage of parsed results.
    :return: an instance of `cls`
    """
    result = dict()
    for item in _parse_list_header(value):
        if '=' not in item:
            result[item] = None
            continue
        name, value = item.split('=', 1)
        if value[:1] == value[-1:] == '"':
            value = unquote_header_value(value[1:-1])
        result[name] = value
    return result


def parse_authorization_header(auth):
    if not auth:
        return

    try:
        auth_scheme, auth_info = auth.split(None, 1)
        auth_scheme = auth_scheme.lower()
    except ValueError:
        return

    result = parse_dict_header(auth_info)
    result['scheme'] = auth_scheme
    return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import pkg_resources

from ava.runtime import environ, settings


logger = logging.getLogger(__name__)


class PackageManager(object):
    """
    Add packages in egg format at runtime.
    """
    def __init__(self, pkgs_dir=environ.pkgs_dir()):
        self.pkgs_dir = pkgs_dir

    def find_packages(self, add_to_path=True):
        """
        Extends sys.path at runtime to include distributions in user's home directory.
        """

        #logger.debug("sys.path(before): ", sys.path)

        logger.debug("Packages Directory: %s", self.pkgs_dir)
        distributions, errors = pkg_resources.working_set.find_plugins(
            pkg_resources.Environment([self.pkgs_dir])
        )

        if len(distributions) > 0:
            logger.debug("Found %d extension package(s).", len(distributions))
            #map(pkg_resources.working_set.add, distributions)  # add plugins+libs to sys.path

            if not add_to_path:
                return

            for it in distributions:
                pkg_resources.working_set.add(it)
                logger.debug("package added: %s", it.project_name)

            logger.error("Couldn't load: %r", errors)        # display errors
        else:
            logger.debug("No extension package found.")




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import sys

"""
For runtime compatibility.
"""

if sys.version_info < (2, 7):
    raise Exception('Hub requires Python 2.7 or higher.')


PY3 = sys.version_info[0] == 3
JYTHON = sys.platform.startswith('java')
PYPY = hasattr(sys, 'pypy_version_info')


if PY3:
    string_types = str,
    text_type = str
else:
    string_types = basestring,
    text_type = unicode


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import six
import hashlib
import pyscrypt
import base58
import hashlib
#from Crypto.Hash import RIPEMD
import libnacl.public
import libnacl.secret

# The public key used for verifying peer tokens which are the identify certificates.
TOKEN_PUBLIC_KEY = ''

XID_PREFIX = b'\x5f'
KEY_PREFIX = b'\xba'  # prefix for public key string
SECRET_PREFIX = b'\xff'  # prefix for secret key string
FINGER_PREFIX = b'\xef'


def generate_keypair(sk=None):
    """
    Generate a random key pair.
    :return:
    """
    if sk:
        keypair = libnacl.public.SecretKey(sk=sk)
    else:
        keypair = libnacl.public.SecretKey()

    return keypair.pk, keypair.sk


def generate_symmetric_key():
    """
    Generate a random key for symmetric encryption.

    :return:
    """
    box = libnacl.secret.SecretBox()
    return box.sk


def key_to_xid(key):
    """
    Generate an address from a public key.
    :param key: the 32-byte byte string.
    :return:
    """
    sha256 = hashlib.sha256()
    sha256.update(XID_PREFIX)
    sha256.update(key)
    sum = sha256.digest()
    addr = XID_PREFIX + key + sum[:2]

    return base58.b58encode(addr)


def xid_to_key(xid):
    """
    Retrieve the key from an XID.

    :param xid:
    :return:
    """
    val = base58.b58decode(xid)

    if len(val) != 35:
        return None

    prefix = val[0]
    if prefix != XID_PREFIX:
        return None

    key = val[1:33]
    sha256 = hashlib.sha256()
    sha256.update(XID_PREFIX)
    sha256.update(key)
    s = sha256.digest()

    if val[-2:] != s[:2]:
        return None

    return key


def validate_xid(addr):
    """
    Check if the provided address is valid or not.

    :param addr: address in base58 string.
    :return:
    """
    val = base58.b58decode(addr)
    #assert len(val) == 35

    if len(val) != 35:
        return False

    prefix = val[0]
    if prefix != XID_PREFIX:
        return False

    sha256 = hashlib.sha256()
    sha256.update(XID_PREFIX)
    sha256.update(val[1:33])
    s = sha256.digest()

    if val[-2:] != s[:2]:
        return False

    return True


def key_to_fingerprint(key):
    """
    Fingerprint is used to uniquely identify an entity without exposing extra information.
    :param key: the public key
    :return:
    """
    #assert len(key) == 32

    sha256 = hashlib.sha256()
    sha256.update(key)
    key_hash = sha256.digest()

    #print('%r' % key_hash)

    #assert len(key_hash) == 32


    ripemd = hashlib.new('ripemd')
    ripemd.update(key_hash)
    key_hash = ripemd.digest()

    #assert len(key_hash) == 20

    sha256 = hashlib.sha256()
    sha256.update(FINGER_PREFIX)
    sha256.update(key_hash)
    checksum = sha256.digest()
    result = FINGER_PREFIX+key_hash+checksum[:4]
    return base58.b58encode(result)


def validate_fingerprint(fingerprint):
    val = base58.b58decode(fingerprint)
    #print("Length: %d" % len(val) )
    if len(val) != 25:
        return False

    prefix = val[0]
    if prefix != FINGER_PREFIX:
        return False

    key_hash = val[1:21]
    sha256 = hashlib.sha256()
    sha256.update(FINGER_PREFIX)
    sha256.update(key_hash)
    given_checksum = sha256.digest()[:4]
    expected_checksum = val[-4:]
    if expected_checksum != given_checksum:
        return False

    return True


def xid_to_fingerprint(address):
    key = xid_to_key(address)
    if key:
        return key_to_fingerprint(key)
    else:
        return None


def key_to_peer_id(key):
    sha256 = hashlib.sha256()
    sha256.update(key)
    key_hash = sha256.digest()

    #print('%r' % key_hash)

    #assert len(key_hash) == 32

    ripemd = hashlib.new('ripemd')
    ripemd.update(key_hash)
    return ripemd.digest()


def calc_session_key(pk, sk):
    """
    Calculate the session key from a public key and secret key.

    :param pk: the counterpart's public key.
    :param sk: the own private key
    :return:
    """
    return libnacl.crypto_box_beforenm(pk, sk)


def secret_key_encrypt(sk, plaindata):
    """
    Encrypt a chunk of data with secret key symmetric encryption.

    :param sk:
    :param plaindata:
    :return:
    """
    box = libnacl.secret.SecretBox(sk)
    return box.encrypt(plaindata)


def secret_key_decrypt(sk, encrypted):
    box = libnacl.secret.SecretBox(sk)
    return box.decrypt(encrypted)


def secret_key_box(sk):
    return libnacl.secret.SecretBox(sk)


def public_key_encrypt(sender_sk, receiver_pk, plaintext):
    sender_box = libnacl.public.Box(sender_sk, receiver_pk)
    return sender_box.encrypt(plaintext)


def public_key_decrypt(receiver_sk, sender_pk, ciphertext):
    receiver_box = libnacl.public.Box(receiver_sk, sender_pk)
    return receiver_box.decrypt(ciphertext)


def public_key_box(sk, pk):
    return libnacl.public.Box(sk, pk)


def derive_secret_key(password=None, salt=None):
    """
    Derives secret key from password and salt combination with Scrypt.

    :param password:
    :param salt:
    :return:
    """
    hash = pyscrypt.hash(password=password,
                         salt=salt,
                         N=1024,
                         r=1,
                         p=1,
                         dkLen=32)

    return hash


def secret_key_to_xid(sk):
    """
    Derives XID from a secret key.

    :param sk:
    :return:
    """
    keyobj = libnacl.public.SecretKey(sk=sk)
    pk = keyobj.pk
    return key_to_xid(pk)


def secret_to_string(key):
    """
    Converts a secret key to its string form
    :param key: the secret key
    :return: a string equivalent to the secret key
    """
    sha256 = hashlib.sha256()
    sha256.update(SECRET_PREFIX)
    sha256.update(key)
    sum = sha256.digest()
    result = SECRET_PREFIX + key + sum[:2]

    return base58.b58encode(result)


def string_to_secret(sk_str):
    val = base58.b58decode(sk_str)

    if len(val) != 35:
        return None

    prefix = val[0]
    if prefix != SECRET_PREFIX:
        return None

    key = val[1:33]
    sha256 = hashlib.sha256()
    sha256.update(SECRET_PREFIX)
    sha256.update(key)
    s = sha256.digest()

    if val[-2:] != s[:2]:
        return None

    return key


def validate_secret_string(sk_str):
    """
    Validates a given public key string.
    :param sk_str:
    :return:
    """
    val = base58.b58decode(sk_str)

    if len(val) != 35:
        return False

    prefix = val[0]
    if prefix != SECRET_PREFIX:
        return False

    sha256 = hashlib.sha256()
    sha256.update(SECRET_PREFIX)
    sha256.update(val[1:33])
    s = sha256.digest()

    if val[-2:] != s[:2]:
        return False

    return True


def key_to_string(key):
    """
    Converts a public key to its string form
    :param key: the public key
    :return:
    """
    sha256 = hashlib.sha256()
    sha256.update(KEY_PREFIX)
    sha256.update(key)
    sum = sha256.digest()
    result = KEY_PREFIX + key + sum[:2]

    return base58.b58encode(result)


def string_to_key(pk_str):
    val = base58.b58decode(pk_str)

    if len(val) != 35:
        return None

    prefix = val[0]
    if prefix != KEY_PREFIX:
        return None

    key = val[1:33]
    sha256 = hashlib.sha256()
    sha256.update(KEY_PREFIX)
    sha256.update(key)
    s = sha256.digest()

    if val[-2:] != s[:2]:
        return None

    return key


def validate_key_string(pk_str):
    """
    Validates a given public key string.
    :param sk_str:
    :return:
    """
    val = base58.b58decode(pk_str)
    #assert len(val) == 35

    if len(val) != 35:
        return False

    prefix = val[0]
    if prefix != KEY_PREFIX:
        return False

    sha256 = hashlib.sha256()
    sha256.update(KEY_PREFIX)
    sha256.update(val[1:33])
    s = sha256.digest()

    if val[-2:] != s[:2]:
        return False

    return True



# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import

import time

from gi.repository import Gtk
from gi.repository import AppIndicator3 as appindicator
from ava.shell.base import ShellBase, STR_OPEN_WEBFRONT, STR_EXIT
from ava.shell import resource_path
from ava.runtime import environ

import click

class StatusIcon(object):
    def __init__(self, shell):
        self.shell = shell
        self.ind = appindicator.Indicator.new("Ava-indicator",
                                           resource_path("res/eavatar.png"),
                                           appindicator.IndicatorCategory.APPLICATION_STATUS)
        self.ind.set_icon_theme_path(resource_path('res/'))

        #self.ind = appindicator.Indicator.new_with_path ("EAvatar-indicator", "eavatar.png",
        #        appindicator.IndicatorCategory.APPLICATION_STATUS,
        #        resource_path('res'))
        self.ind.set_status(appindicator.IndicatorStatus.ACTIVE)
        self.ind.set_attention_icon("eavatar.png")

        self.menu_setup()
        self.ind.set_menu(self.menu)

    def menu_setup(self):
        self.menu = Gtk.Menu()

        self.open_item = Gtk.MenuItem.new_with_label(STR_OPEN_WEBFRONT)
        self.open_item.connect("activate", self.on_open_frame)
        self.open_item.show()
        self.quit_item = Gtk.MenuItem.new_with_label(STR_EXIT)
        self.quit_item.connect("activate", self.quit)
        self.quit_item.show()

        self.menu.append(self.open_item)
        self.menu.append(Gtk.SeparatorMenuItem.new())
        self.menu.append(self.quit_item)

    def on_open_frame(self, sender):
        self.shell.open_main_ui()

    def on_open_folder(self, sender):
        click.launch(environ.pod_dir())

    def quit(self, widget):
        self.ind.set_status(appindicator.IndicatorStatus.PASSIVE)
        Gtk.main_quit()


def yieldToOthers():
    time.sleep(0.1)
    #GLib.timeout_add(200, yieldToOthers)
    return True


class Shell(ShellBase):
    def __init__(self):
        super(Shell, self).__init__()
        self.statusIcon = StatusIcon(self)

    def run(self):
        #gtk.timeout_add(200, self._on_idle)
        #GLib.timeout_add(200, yieldToOthers, priority=GLib.PRIORITY_DEFAULT)
        #GObject.idle_add(self._on_idle)
        Gtk.main()


if __name__ == '__main__':
    shell = Shell()
    shell.run()




# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from ava import __version__

setup(
    name="ava",
    version=__version__,
    description="EAvatar Ava - A versatile agent.",
    # package_dir={'': ''},
    packages=find_packages(exclude=['tests']),
    include_package_data=True,

    install_requires=['backports.ssl-match-hostname',
                      'base58',
                      'bottle',
                      'click',
                      'gevent',
                      'libnacl',
                      'lmdb',
                      'msgpack-python',
                      'PyDispatcher',
                      'PyYAML',
                      'pyscrypt',
                      'requests',
                      'six',
                      'ujson',
                      'ws4py',
                      'wsaccel'],

    test_suite='nose.collector',
    zip_safe=False,

    entry_points={
        'console_scripts': [
            'ava = ava.launcher:main',
        ],
    },

    author="Sam Kuo",
    author_email="sam@eavatar.com",
    url="http://www.eavatar.com",

)


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import time
import unittest

import mock
from ava.shell.base import Supervisor


def server_process1():
    time.sleep(2)


class SupervisorTest(unittest.TestCase):

    @mock.patch('ava.shell.base.multiprocessing.Process')
    def test_supervisor(self, mock_process_class):
        mp = mock_process_class.return_value
        mp.exitcode = None

        supervisor = Supervisor(target=server_process1)
        self.assertFalse(supervisor.is_server_running())

        supervisor.start_server()
        self.assertTrue(mock_process_class.called)
        self.assertTrue(mp.start.called)
        self.assertTrue(supervisor.is_server_running())

        supervisor.stop_server()
        self.assertTrue(mp.terminate.called)
        self.assertFalse(supervisor.is_server_running())

    @mock.patch('ava.shell.base.multiprocessing.Process')
    def test_supervisor_with_server_abnormally_exit(self, mock_process_class):
        mp = mock_process_class.return_value
        mp.exitcode = None

        supervisor = Supervisor(target=server_process1)
        self.assertFalse(supervisor.is_server_running())

        supervisor.start_server()
        self.assertTrue(mock_process_class.called)
        self.assertTrue(mp.start.called)
        self.assertTrue(supervisor.is_server_running())

        # emulate server process exits abnormally.
        mp.exitcode = -1
        self.assertTrue(supervisor.health_check())
        self.assertEqual(mp.start.call_count, 2)

        # exitcode = 1 indicates that server requested to restart.
        mp.exitcode = 1
        self.assertTrue(supervisor.health_check())
        self.assertEqual(mp.start.call_count, 3)

        # in case that restarting too many times, supervisor should give up.
        for i in range(5):
            mp.exitcode = -1
            supervisor.health_check()

        self.assertFalse(supervisor.health_check())

        supervisor.stop_server()
        self.assertFalse(mp.terminate.called)
        self.assertFalse(supervisor.is_server_running())

    @mock.patch('ava.shell.base.multiprocessing.Process')
    def test_supervisor_with_server_normally_exit(self, mock_process_class):
        mp = mock_process_class.return_value
        mp.exitcode = None

        supervisor = Supervisor(target=server_process1)
        self.assertFalse(supervisor.is_server_running())

        supervisor.start_server()
        self.assertTrue(mock_process_class.called)
        self.assertTrue(mp.start.called)
        self.assertTrue(supervisor.is_server_running())

        # emulate server process exits normally.
        mp.exitcode = 0
        self.assertFalse(supervisor.health_check())
        self.assertEqual(mp.start.call_count, 1)

        supervisor.stop_server()
        self.assertFalse(mp.terminate.called)
        self.assertFalse(supervisor.is_server_running())



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import unittest
from io import BytesIO
from msgpack import Packer, Unpacker, packb, unpackb


class Data(dict):
    def __init__(self, name):
        super(Data, self).__init__()
        self['name'] = name

    def __getattr__(self, item):
        return self[item]


class MsgPackTest(unittest.TestCase):

    def test_packer_unpacker(self):
        buf = BytesIO()
        packer = Packer()
        buf.write(packer.pack(1))
        buf.write(packer.pack('2'))
        buf.write(packer.pack({}))
        buf.seek(0)
        unpacker = Unpacker(buf)
        v1 = unpacker.unpack()
        self.assertEqual(1, v1)

        v2 = unpacker.unpack()
        self.assertEqual('2', v2)

        v3 = unpacker.unpack()
        self.assertTrue(isinstance(v3, dict))

    def test_pack_dict(self):
        data = {'name': 'test1'}
        packed = packb(data)
        unpacked = unpackb(packed)
        self.assertTrue(isinstance(unpacked, dict))
        self.assertEqual('test1', unpacked['name'])

    def test_pack_tuple(self):
        data = ('a', 1, False)
        packed = packb(data)
        unpacked = unpackb(packed)

        self.assertTrue(isinstance(unpacked, list))
        self.assertEqual(unpacked[0], 'a')
        self.assertEqual(unpacked[1], 1)
        self.assertEqual(unpacked[2], False)

    def test_pack_object(self):
        data = Data('test')
        packed = packb(data)
        unpacked = unpackb(packed)
        self.assertEqual(unpacked['name'], 'test')


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import gevent
from gevent import monkey
monkey.patch_all()

import os
import sys
import signal
import logging
import importlib

from ava.util import crypto
from ava.runtime import settings
from ava.runtime import environ
from ava.runtime import config

from ava.spi import context
from ava.spi.defines import INSTALLED_ENGINES, AVA_AGENT_SECRET, AVA_USER_XID
from ava.spi.signals import AGENT_STARTED, AGENT_STOPPING

KEYFILE = 'ava-keys.yml'

logger = logging.getLogger(__name__)

__agent = None



def _mygetfilesystemencoding():
    old = sys.getfilesystemencoding

    def inner_func():
        ret = old()
        if ret is None:
            return 'UTF-8'
        else:
            return ret
    return inner_func


def patch_sys_getfilesystemencoding():
    # sys.getfilesystemencoding() always returns None when frozen on Ubuntu systems.
    patched_func = _mygetfilesystemencoding()
    sys.getfilesystemencoding = patched_func


def restart_later():
    if __agent.running:
        logger.warning("Agent not stopped successfully!")
    sys.exit(1)


def signal_handler(signum = None, frame = None):
    logger.debug("Received HUP signal, requests the shell to restart.")
    global __agent
    if __agent:
        __agent._stop_engines()
    sys.exit(1)


def load_class(full_class_string):
    """
    dynamically load a class from a string. e.g. 'a.b.package:classname'
    """

    class_data = full_class_string.split(":")
    module_path = class_data[0]
    class_name = class_data[1]

    module = importlib.import_module(module_path)
    # Finally, we retrieve the Class
    return class_name, getattr(module, class_name)


class Agent(object):
    def __init__(self):
        logger.debug("Initializing agent...")
        patch_sys_getfilesystemencoding()

        self.running = False
        self.interrupted = False
        self._greenlets = []
        self._context = context.instance(self)
        self._engines = []
        self.__secret = None
        self.xid = None
        self.key = None
        self.user_xid = None

        self.get_security_keys()
        self.save_keys()

        if hasattr(signal, 'SIGHUP'):
            signal.signal(signal.SIGHUP, signal_handler)

    def get_security_keys(self):
        keyfile = os.path.join(environ.conf_dir(), KEYFILE)
        keys = config.load_conf(keyfile)

        self.__secret = keys.get('secret')
        self.user_xid = keys.get('user')
        self.__secret = os.environ.get(AVA_AGENT_SECRET, self.__secret)
        self.user_xid = os.environ.get(AVA_USER_XID, self.user_xid)

        if self.user_xid is None and not settings['debug']:
            logger.error('No User XID is specified!')
            raise SystemExit(2)

        if self.__secret:
            self.__secret = crypto.string_to_secret(self.__secret)
            pk, sk = crypto.generate_keypair(sk=self.__secret)
            self.key = pk
        else:
            logger.debug("No secret key is given, generating one...")
            pk, sk = crypto.generate_keypair()
            self.__secret = sk
            self.key = pk

        self.xid = crypto.key_to_xid(self.key)
        logger.debug("The agent's XID: %s", self.xid)

        if not self.user_xid and settings['debug']:
            logger.debug("User XID not given via environment variable. " +
                         "Generating one...")
            self.user_xid = \
                b'AYPwK3c3VK7ZdBvKfcbV5EmmCZ8zSb9viZ288gKFBFuE92jE'
        logger.debug("The agent's user XID: %s", self.user_xid)

    def save_keys(self):
        keyfile = os.path.join(environ.conf_dir(), KEYFILE)
        keys = {
            b'secret': crypto.secret_to_string(self.__secret),
            b'user': self.user_xid,
        }

        config.save_conf(keyfile, keys)

    def add_child_greenlet(self, child):
        self._greenlets.append(child)

    def _start_engines(self):
        for it in INSTALLED_ENGINES:
            logger.debug("Loading engine: %s", it)
            name, engine_cls = load_class(it)
            engine = engine_cls()

            self._context.bind(name, engine)
            self._engines.append(self._context[name])

        logger.debug("Starting engines...")
        for it in self._engines:
            it.start(self._context)

        self._context.send(signal=AGENT_STARTED, sender=self)

    def _stop_engines(self):
        self._context.send(signal=AGENT_STOPPING, sender=self)

        for it in reversed(self._engines):
            try:
                it.stop(self._context)
            except:
                logger.warning("Error while stopping %s", it.__name__)

    def context(self):
        return self._context

    def run(self):
        logger.debug("Starting agent...")

        self._start_engines()

        self.running = True
        logger.debug("Agent started.")

        while not self.interrupted:
            try:
                gevent.joinall(self._greenlets, timeout=1)
            except KeyboardInterrupt:
                logger.debug("Interrupted.")
                break

        # stop engines in reverse order.
        self._stop_engines()

        gevent.killall(self._greenlets, timeout=1)

        self.running = False

        logger.debug("Agent stopped.")


def start_agent():
    global __agent
    __agent = Agent()
    __agent.run()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')
    start_agent()




# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import unittest
import libnacl.public
import libnacl.secret
import libnacl.utils
import pyscrypt
import base58
import binascii
from ava.util import crypto


class CryptoTest(unittest.TestCase):

    def test_generate_keypair(self):
        bob = libnacl.public.SecretKey()
        assert bob is not None

        self.assertEqual(len(bob.sk), 32)
        sk = base58.b58encode(bob.sk)
        pk = base58.b58encode(bob.pk)
        xid = crypto.key_to_xid(bob.pk)

        print("Public key: ", pk, ", len: ", len(pk))
        print("Secret key: ", sk, ", len: ", len(sk))
        print("XID: ", xid, ", len: ", len(xid))

    def test_public_key_encryption(self):
        msg = b'You\'ve got two empty halves of coconut and you\'re bangin\' \'em together.'
        bob = libnacl.public.SecretKey()
        alice = libnacl.public.SecretKey()
        bob_box = libnacl.public.Box(bob.sk, alice.pk)
        alice_box = libnacl.public.Box(alice.sk, bob.pk)
        bob_ctxt = bob_box.encrypt(msg)
        bclear = alice_box.decrypt(bob_ctxt)
        alice_ctxt = alice_box.encrypt(msg)
        aclear = alice_box.decrypt(alice_ctxt)

        self.assertEqual(bclear, aclear)
        self.assertEqual(bclear, msg)

    def test_public_key_encryption_util(self):
        alice = crypto.generate_keypair()
        bob = crypto.generate_keypair()

        ciphertext = crypto.public_key_encrypt(alice[1], bob[0], "hello")
        plaintext = crypto.public_key_decrypt(bob[1], alice[0], ciphertext)
        self.assertEqual("hello", plaintext)

    def test_secret_key_encryption(self):
        msg = b'But then of course African swallows are not migratory.'
        # Create a SecretBox object, if not passed in the secret key is
        # Generated purely from random data
        box = libnacl.secret.SecretBox()
        # Messages can now be safely encrypted
        ctxt = box.encrypt(msg)
        # An addition box can be created from the original box secret key
        box2 = libnacl.secret.SecretBox(box.sk)
        # Messages can now be easily encrypted and decrypted
        clear1 = box.decrypt(ctxt)
        clear2 = box2.decrypt(ctxt)
        ctxt2 = box2.encrypt(msg)
        clear3 = box.decrypt(ctxt2)

        self.assertEqual(clear1, clear2)
        self.assertEqual(clear2, clear3)
        self.assertEqual(msg, clear1)

    def test_nonce_generation(self):

        prev_nonce = libnacl.utils.rand_nonce()
        for i in xrange(100):
            nonce = libnacl.utils.rand_nonce()
            self.assertNotEqual(prev_nonce, nonce)
            prev_nonce = nonce

    def test_demo_user_keys(self):
        sk_base58 = b"29oiwbqkhLGBuX5teL5d2vsiJ3EXk3dpyBiPwA7W9DJG"
        sk_decoded = base58.b58decode(sk_base58)
        hash = pyscrypt.hash(password=b"demouser",
                             salt=b"demouser",
                             N=1024,
                             r=1,
                             p=1,
                             dkLen=32)

        sk = base58.b58encode(hash)
        hex_sk = binascii.b2a_hex(hash)
        print('Secret Key:', sk, 'length: ', len(sk))
        self.assertEqual(sk_base58, sk)
        self.assertEqual(sk_decoded, hash)

        #print(sk)
        keypair = libnacl.public.SecretKey(hash)

        # 2ipFYsqXnrw4Mt2RUWzEQntAH1FEFB8R52rAT3eExn9S
        pk_base58 = base58.b58encode(keypair.pk)
        pk_decoded = base58.b58decode(pk_base58)

        self.assertEqual(pk_decoded, keypair.pk)
        print('Public Key:', pk_base58, 'length: ', len(pk_base58))

        print("XID: ", crypto.key_to_xid(keypair.pk))

    def test_calc_shared_key_via_two_key_pairs(self):
        bob = libnacl.public.SecretKey()
        alice = libnacl.public.SecretKey()
        alice_key = libnacl.crypto_box_beforenm(bob.pk, alice.sk)
        bob_key = libnacl.crypto_box_beforenm(alice.pk, bob.sk)
        #print(base58.b58encode(bob_key))
        self.assertEqual(bob_key, alice_key)

    def test_key_to_fingerprint(self):
        bob = libnacl.public.SecretKey()
        assert bob is not None

        pk = bob.pk

        fp = crypto.key_to_fingerprint(pk)
        print("FP: ", fp, ", len: ", len(fp))

    def test_convert_key_to_xid_back_and_forth(self):
        keypair = libnacl.public.SecretKey()
        xid = crypto.key_to_xid(keypair.pk)
        print(xid)
        self.assertTrue(crypto.validate_xid(xid))
        pk = crypto.xid_to_key(xid)
        self.assertEqual(keypair.pk, pk)

    def test_convert_key_to_string_back_and_forth(self):
        keypair = libnacl.public.SecretKey()
        pk_str = crypto.key_to_string(keypair.pk)
        # print(pk_str)
        self.assertTrue(crypto.validate_key_string(pk_str))
        pk = crypto.string_to_key(pk_str)
        self.assertEqual(keypair.pk, pk)

    def test_convert_secret_to_string_back_and_forth(self):
        keypair = libnacl.public.SecretKey()
        sk_str = crypto.secret_to_string(keypair.sk)
        # print(sk_str)
        self.assertTrue(crypto.validate_secret_string(sk_str))
        sk = crypto.string_to_secret(sk_str)
        self.assertEqual(keypair.sk, sk)


# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import click
from .cli import cli
from ava.runtime import settings


@cli.command()
@click.argument("app")
def launch(app):
    """ Launch web application.
    """
    click.launch('http://127.0.0.1:%s' % settings['webfront']['listen_port'])


# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import logging

POD_DIR_ENV = 'AVA_POD'

POD_DIR_NAME = u'pod'
PKGS_DIR_NAME = u'pkgs'
LOGS_DIR_NAME = u'logs'
DATA_DIR_NAME = u'data'
CONF_DIR_NAME = u'conf'
MODS_DIR_NAME = u'mods'


class Environment(object):
    """
    Encapsulates the runtime environment.
    """
    def __init__(self):
        self.logger = logging.getLogger(__name__)

        # Determines the location of the base directory which contains files
        # for a specific user.
        # This script assumes it is located at 'ava/runtime' sub-directory.
        from ava.util import base_path

        self.base_dir = base_path()

        # Determines the location of the home directory.

        self.pod_dir = os.path.join(self.base_dir, POD_DIR_NAME)
        self.pod_dir = os.environ.get(POD_DIR_ENV, self.pod_dir)
        self.pod_dir = os.path.abspath(self.pod_dir)
        if not os.path.isdir(self.pod_dir):
            self.logger.error("Invalid pod folder: %s" % self.pod_dir)
            sys.exit(-1)

        self.conf_dir = os.path.join(self.pod_dir, CONF_DIR_NAME)
        self.pkgs_dir = os.path.join(self.pod_dir, PKGS_DIR_NAME)
        self.data_dir = os.path.join(self.pod_dir, DATA_DIR_NAME)
        self.logs_dir = os.path.join(self.pod_dir, LOGS_DIR_NAME)
        self.mods_dir = os.path.join(self.pod_dir, MODS_DIR_NAME)

        # Flag indicating if the runtime is launched by a shell.
        self.has_shell = False


# The global environment.
_environ = None


def get_environ():
    global _environ

    if _environ is None:
        _environ = Environment()

    return _environ


def base_dir():
    """
    Gets the base directory.
    :return:
    """
    return get_environ().base_dir


def pod_dir():
    """
    Gets the home directory.
    :return:
    """
    return get_environ().pod_dir


def conf_dir():
    """
    Gets the path for configuration files.

    :return: The configuration path.
    """
    return get_environ().conf_dir


def data_dir():
    """
    Gets the path for data files.

    :return: The path.
    """
    return get_environ().data_dir


def logs_dir():
    """
    Gets the path for log files.

    :return: The path.
    """
    return get_environ().logs_dir


def pkgs_dir():
    """
    Gets the path for packages files.

    :return: The path.
    """
    return get_environ().pkgs_dir


def mods_dir():
    return get_environ().mods_dir
